from plone.app.testing import setRoles
from plone.app.testing import TEST_USER_ID
from plone.app.z3cform.tests.layer import PAZ3CForm_INTEGRATION_TESTING
from plone.app.z3cform.widgets.querystring import get_querystring_options
from plone.app.z3cform.widgets.relateditems import get_relateditems_options
from plone.app.z3cform.widgets.richtext import get_tinymce_options

import unittest


class TestQueryStringOptions(unittest.TestCase):
    layer = PAZ3CForm_INTEGRATION_TESTING

    def setUp(self):
        setRoles(self.layer["portal"], TEST_USER_ID, ["Contributor"])

    def test__query_string_options(self):
        """Test query string options on root:
        All URLs and paths equal root url and path,
        no favorites
        """

        portal = self.layer["portal"]
        options = get_querystring_options(portal, "@@qsOptions")

        # Test base options
        self.assertEqual(options["indexOptionsUrl"], "http://nohost/plone/@@qsOptions")

        self.assertEqual(
            options["previewCountURL"],
            "http://nohost/plone/@@querybuildernumberofresults",
        )

        self.assertEqual(
            options["previewURL"], "http://nohost/plone/@@querybuilder_html_results"
        )

        # Test options of the AJAX select widget
        self.assertEqual(options["patternAjaxSelectOptions"]["separator"], ";")

        # Test options of the date picker
        self.assertEqual(
            options["patternDateOptions"],
            {
                "behavior": "native",
                "week-numbers": "show",
                "first-day": 0,
                "today": "Today",
                "clear": "Clear",
            },
        )

        # Test options of the related items widget
        self.assertEqual(options["patternRelateditemsOptions"]["basePath"], "/plone")
        self.assertTrue("recentlyUsed" not in options["patternRelateditemsOptions"])


class TestRelatedItemsOptions(unittest.TestCase):
    layer = PAZ3CForm_INTEGRATION_TESTING

    def setUp(self):
        setRoles(self.layer["portal"], TEST_USER_ID, ["Contributor"])

    def test__base_relateditems_options(self):
        """Test related items options on root:
        All URLs and paths equal root url and path,
        no favorites
        """

        portal = self.layer["portal"]
        options = get_relateditems_options(
            portal, None, "#!@", "test_vocab", "@@vocab", "testfield"
        )

        # vocab is correctly set
        self.assertTrue(
            "@@vocab?name=test_vocab&field=testfield" in options["vocabularyUrl"]
        )

        # rootUrl contains something
        self.assertTrue(bool(options["rootUrl"]))

        root_path = context_path = "/".join(portal.getPhysicalPath())
        root_url = context_url = portal.absolute_url()

        # context_path contains something, otherwise this test is meaningless
        self.assertTrue(bool(context_path))
        # context_url contains something, otherwise this test is meaningless
        self.assertTrue(bool(context_url))

        self.assertEqual(options["rootUrl"], root_url)

        self.assertEqual(options["rootPath"], root_path)

        self.assertEqual(
            options["vocabularyUrl"],
            root_url + "/@@vocab?name=test_vocab&field=testfield",
        )

        self.assertEqual(options["basePath"], context_path)

        self.assertEqual(options["contextPath"], context_path)

        self.assertEqual(options["separator"], "#!@")

        self.assertTrue("favorites" not in options)

        # Recently used is configured, but off per default.
        self.assertEqual(options["recentlyUsed"], False)
        self.assertEqual(
            options["recentlyUsedKey"],
            "relateditems_recentlyused_testfield_" + TEST_USER_ID,
        )

    def test__subfolder_relateditems_options(self):
        """Test related items options on subfolder:
        Vocab called on root, start path is folder, have favorites.
        """

        portal = self.layer["portal"]
        portal.invokeFactory("Folder", "sub")
        sub = portal.sub
        options = get_relateditems_options(
            sub, None, "#!@", "test_vocab", "@@vocab", "testfield"
        )

        # vocab is correctly set
        self.assertTrue(
            "@@vocab?name=test_vocab&field=testfield" in options["vocabularyUrl"]
        )

        # rootUrl contains something
        self.assertTrue(bool(options["rootUrl"]))

        root_path = "/".join(portal.getPhysicalPath())
        root_url = portal.absolute_url()
        context_path = "/".join(sub.getPhysicalPath())
        context_url = sub.absolute_url()

        # context_path contains something, otherwise this test is meaningless
        self.assertTrue(bool(context_path))
        # context_url contains something, otherwise this test is meaningless
        self.assertTrue(bool(context_url))

        self.assertEqual(options["rootUrl"], root_url)

        self.assertEqual(options["rootPath"], root_path)

        self.assertEqual(
            options["vocabularyUrl"],
            root_url + "/@@vocab?name=test_vocab&field=testfield",
        )

        self.assertEqual(options["basePath"], context_path)

        self.assertEqual(options["contextPath"], context_path)

        self.assertEqual(options["separator"], "#!@")

        self.assertEqual(len(options["favorites"]), 2)

        self.assertEqual(sorted(options["favorites"][0].keys()), ["path", "title"])

    def test__subdocument_relateditems_options(self):
        """Test related items options on subdoc:
        Vocab called on root, start path is root as document is not folderish,
        no favorites.
        """

        portal = self.layer["portal"]
        portal.invokeFactory("Document", "sub")
        sub = portal.sub
        options = get_relateditems_options(
            sub, None, "#!@", "test_vocab", "@@vocab", "testfield"
        )

        # vocab is correctly set
        self.assertTrue(
            "@@vocab?name=test_vocab&field=testfield" in options["vocabularyUrl"]
        )

        # rootUrl contains something
        self.assertTrue(bool(options["rootUrl"]))

        root_path = "/".join(portal.getPhysicalPath())
        root_url = portal.absolute_url()
        context_path = "/".join(sub.getPhysicalPath())
        context_url = sub.absolute_url()

        # context_path contains something, otherwise this test is meaningless
        self.assertTrue(bool(context_path))
        # context_url contains something, otherwise this test is meaningless
        self.assertTrue(bool(context_url))

        self.assertEqual(options["rootUrl"], root_url)

        self.assertEqual(options["rootPath"], root_path)

        self.assertEqual(
            options["vocabularyUrl"],
            root_url + "/@@vocab?name=test_vocab&field=testfield",
        )

        self.assertEqual(options["basePath"], root_path)

        self.assertEqual(options["contextPath"], context_path)

        self.assertEqual(options["separator"], "#!@")

        self.assertTrue("favorites" not in options)


class TestTinyMCEOptions(unittest.TestCase):
    layer = PAZ3CForm_INTEGRATION_TESTING

    def setUp(self):
        setRoles(self.layer["portal"], TEST_USER_ID, ["Contributor"])

    def test__tinymce_options_different_contexts(self):
        """Test if ``get_tinymce_options`` can be called with different
        contexts, including invalid and form contexts.
        """
        request = self.layer["request"]
        portal = self.layer["portal"]
        portal.invokeFactory("Folder", "sub")
        sub = portal.sub

        # TinyMCE on portal context
        options = get_tinymce_options(portal, None, request)
        self.assertEqual(options["relatedItems"]["basePath"], "/plone")

        # TinyMCE on sub folder context
        options = get_tinymce_options(sub, None, request)
        self.assertEqual(options["relatedItems"]["basePath"], "/plone/sub")


class BaseWidgetTests(unittest.TestCase):
    """Tests for plone.app.z3cform.widgets.patterns.BaseWidget."""

    def test_defaults(self):
        from plone.app.z3cform.widgets.patterns import BaseWidget

        widget = BaseWidget("input", "example1")
        self.assertEqual(widget.render(), '<input class="pat-example1"/>')

        self.assertEqual(widget.klass, "pat-example1")

    def test_different_element_tag(self):
        from plone.app.z3cform.widgets.patterns import BaseWidget

        widget = BaseWidget("select", "example1")
        self.assertEqual(widget.render(), '<select class="pat-example1"/>')

        self.assertEqual(widget.klass, "pat-example1")

    def test_setting_patterns_options(self):
        from plone.app.z3cform.widgets.patterns import BaseWidget

        widget = BaseWidget(
            "input",
            "example1",
            pattern_options={
                "option1": "value1",
                "option2": "value2",
            },
        )

        html = widget.render()
        # the order of options is non-deterministic
        result1 = '<input class="pat-example1" data-pat-example1="{&quot;option1&quot;: &quot;value1&quot;, &quot;option2&quot;: &quot;value2&quot;}"/>'  # noqa: E501
        result2 = '<input class="pat-example1" data-pat-example1="{&quot;option2&quot;: &quot;value2&quot;, &quot;option1&quot;: &quot;value1&quot;}"/>'  # noqa: E501
        self.assertIn(html, [result1, result2])


class InputWidgetTests(unittest.TestCase):
    """Tests for plone.app.z3cform.widgets.patterns.InputWidget."""

    def test_defaults(self):
        from plone.app.z3cform.widgets.patterns import InputWidget

        widget = InputWidget("example1", name="example2")

        self.assertEqual(
            widget.render(), '<input class="pat-example1" type="text" name="example2"/>'
        )

        self.assertEqual(widget.type, "text")
        self.assertEqual(widget.value, None)

    def test_set_type_and_value(self):
        from plone.app.z3cform.widgets.patterns import InputWidget

        widget = InputWidget(
            "example1", name="example2", type="email", value="example3"
        )

        self.assertEqual(
            widget.render(),
            '<input class="pat-example1" type="email" '
            'name="example2" value="example3"/>',
        )

        self.assertEqual(widget.type, "email")
        self.assertEqual(widget.value, "example3")

        widget.type = "text"
        widget.value = "example4"
        self.assertEqual(
            widget.render(),
            '<input class="pat-example1" type="text" '
            'name="example2" value="example4"/>',
        )

        self.assertEqual(widget.type, "text")
        self.assertEqual(widget.value, "example4")

        del widget.type
        del widget.value
        self.assertEqual(
            widget.render(), '<input class="pat-example1" name="example2"/>'
        )

        self.assertEqual(widget.type, None)
        self.assertEqual(widget.value, None)


class SelectWidgetTests(unittest.TestCase):
    """Tests for plone.app.z3cform.widgets.patterns.SelectWidget."""

    def test_defaults(self):
        from plone.app.z3cform.widgets.patterns import SelectWidget

        widget = SelectWidget("example1", name="example2")

        self.assertEqual(
            widget.render(), '<select class="pat-example1" name="example2"></select>'
        )
        self.assertEqual(list(widget.items), [])
        self.assertEqual(widget.value, [])

    def test_set_items_and_value(self):
        from plone.app.z3cform.widgets.patterns import SelectWidget

        items = [
            ("token1", "value1"),
            ("token2", "value2"),
            ("token3", "value3"),
        ]
        widget = SelectWidget("example1", name="example2", value="token2", items=items)

        self.assertEqual(
            widget.render(),
            '<select class="pat-example1" name="example2">'
            '<option value="token1">value1</option>'
            '<option value="token2" selected="selected">value2</option>'
            '<option value="token3">value3</option>'
            "</select>",
        )

        self.assertEqual(list(widget.items), items)
        self.assertEqual(widget.value, ["token2"])

        widget.value = "token1"
        self.assertEqual(
            widget.render(),
            '<select class="pat-example1" name="example2">'
            '<option value="token1" selected="selected">value1</option>'
            '<option value="token2">value2</option>'
            '<option value="token3">value3</option>'
            "</select>",
        )

        self.assertEqual(list(widget.items), items)
        self.assertEqual(widget.value, ["token1"])

        del widget.value
        self.assertEqual(
            widget.render(),
            '<select class="pat-example1" name="example2">'
            '<option value="token1">value1</option>'
            '<option value="token2">value2</option>'
            '<option value="token3">value3</option>'
            "</select>",
        )

        del widget.items
        self.assertEqual(
            widget.render(), '<select class="pat-example1" name="example2"></select>'
        )

    def test_multiple(self):
        from plone.app.z3cform.widgets.patterns import SelectWidget

        items = [
            ("token1", "value1"),
            ("token2", "value2"),
            ("token3", "value3"),
        ]
        widget = SelectWidget(
            "example1",
            name="example2",
            value=["token2"],
            items=items,
            multiple=True,
        )

        self.assertEqual(
            widget.render(),
            '<select class="pat-example1" multiple="multiple" name="example2">'
            '<option value="token1">value1</option>'
            '<option value="token2" selected="selected">value2</option>'
            '<option value="token3">value3</option>'
            "</select>",
        )

        self.assertEqual(list(widget.items), items)
        self.assertEqual(widget.value, ["token2"])

        widget.value = ["token1", "token2"]
        self.assertEqual(
            widget.render(),
            '<select class="pat-example1" multiple="multiple" name="example2">'
            '<option value="token1" selected="selected">value1</option>'
            '<option value="token2" selected="selected">value2</option>'
            '<option value="token3">value3</option>'
            "</select>",
        )

        self.assertEqual(list(widget.items), items)
        self.assertEqual(widget.value, ["token1", "token2"])

        del widget.value
        self.assertEqual(
            widget.render(),
            '<select class="pat-example1" multiple="multiple" name="example2">'
            '<option value="token1">value1</option>'
            '<option value="token2">value2</option>'
            '<option value="token3">value3</option>'
            "</select>",
        )

        del widget.items
        self.assertEqual(
            widget.render(),
            '<select class="pat-example1" multiple="multiple" '
            'name="example2"></select>',
        )


class TextareaWidgetTests(unittest.TestCase):
    """Tests for plone.app.z3cform.widgets.patterns.TextareaWidget."""

    def test_defaults(self):
        from plone.app.z3cform.widgets.patterns import TextareaWidget

        widget = TextareaWidget("example1", name="example2")
        self.assertEqual(
            widget.render(),
            '<textarea class="pat-example1" name="example2"></textarea>',
        )

        self.assertEqual(widget.name, "example2")
        self.assertEqual(widget.klass, "pat-example1")
        self.assertEqual(widget.value, "")

    def test_setting_patterns_options(self):
        from plone.app.z3cform.widgets.patterns import TextareaWidget

        widget = TextareaWidget(
            "example1",
            name="example2",
            pattern_options={
                "option1": "value1",
                "option2": "value2",
            },
        )

        html = widget.render()
        # the order of options is non-deterministic
        result1 = '<textarea class="pat-example1" name="example2" data-pat-example1="{&quot;option1&quot;: &quot;value1&quot;, &quot;option2&quot;: &quot;value2&quot;}"></textarea>'  # noqa: E501
        result2 = '<textarea class="pat-example1" name="example2" data-pat-example1="{&quot;option2&quot;: &quot;value2&quot;, &quot;option1&quot;: &quot;value1&quot;}"></textarea>'  # noqa: E501
        self.assertIn(html, [result1, result2])

    def test_set_value(self):
        from plone.app.z3cform.widgets.patterns import TextareaWidget

        widget = TextareaWidget("example1", name="example2", value="example3")
        self.assertEqual(
            widget.render(),
            '<textarea class="pat-example1" name="example2">' "example3" "</textarea>",
        )

        self.assertEqual(widget.value, "example3")

        widget.value = "example4"
        self.assertEqual(
            widget.render(),
            '<textarea class="pat-example1" name="example2">' "example4" "</textarea>",
        )

        del widget.value
        self.assertEqual(
            widget.render(),
            '<textarea class="pat-example1" name="example2"></textarea>',
        )

    def test_can_not_change_element_tag(self):
        from plone.app.z3cform.widgets.patterns import TextareaWidget

        self.assertRaises(
            TypeError, TextareaWidget, "example1", el="input", name="example2"
        )



from Acquisition import aq_base
from plone.i18n.normalizer.interfaces import IIDNormalizer
from Products.Five import BrowserView
from zope.component import queryUtility
from zope.i18n import translate
from zope.i18nmessageid import Message

import json


class InlineValidationView(BrowserView):
    """Validate a form and return the error message for a particular field as JSON."""

    def __call__(self, fname=None, fset=None):
        self.request.response.setHeader("Content-Type", "application/json")

        res = {"errmsg": ""}

        if fname is None:
            return json.dumps(res)

        try:
            form = aq_base(self.context).form_instance
        except AttributeError:
            form = self.context
        try:
            aq_base(form).update()
        except (AttributeError, TypeError):
            return json.dumps(res)

        if getattr(form, "extractData", None):
            data, errors = form.extractData()
        else:
            return json.dumps(res)

        # if we validate a field in a group we operate on the group
        if fset is not None:
            try:
                fset = int(fset)  # integer-indexed fieldset names
                form = form.groups[fset]
            except (ValueError, TypeError):
                # try to match fieldset on group name
                def _name(group):
                    return getattr(group, "__name__", group.label)

                normalize = queryUtility(IIDNormalizer).normalize
                group_match = list(
                    filter(
                        lambda group: normalize(_name(group)) == fset,
                        form.groups,
                    )
                )
                if not group_match:
                    raise ValueError("Fieldset specified, but not found.")
                form = group_match[0]

        index = len(form.prefix) + len(form.widgets.prefix)
        raw_fname = fname[index:]
        validationError = None
        for error in errors:
            if error.widget == form.widgets.get(raw_fname, None):
                validationError = error.message
                break

        if isinstance(validationError, Message):
            validationError = translate(validationError, context=self.request)

        res["errmsg"] = validationError or ""
        return json.dumps(res)



import unittest


class TestUnitCallCallables(unittest.TestCase):
    def test_simple(self):
        from plone.app.z3cform.utils import call_callables

        test_in = 1
        test_compare = 1
        test_out = call_callables(
            test_in,
            "funny return value",
        )
        self.assertEqual(test_out, test_compare)

    def test_simple_function(self):
        from plone.app.z3cform.utils import call_callables

        def test_in(x):
            return x

        test_compare = "funny return value"
        test_out = call_callables(
            test_in,
            "funny return value",
        )
        self.assertEqual(test_out, test_compare)

    def test_list(self):
        from plone.app.z3cform.utils import call_callables

        test_in = [1, 2, 3, lambda x: x]
        test_compare = [1, 2, 3, "funny return value"]
        test_out = call_callables(
            test_in,
            "funny return value",
        )
        self.assertEqual(test_out, test_compare)

    def test_tuple(self):
        from plone.app.z3cform.utils import call_callables

        test_in = (1, 2, 3, lambda x: x)
        test_compare = (1, 2, 3, "funny return value")
        test_out = call_callables(
            test_in,
            "funny return value",
        )
        self.assertEqual(test_out, test_compare)

    def test_complex(self):
        from plone.app.z3cform.utils import call_callables

        test_in = {
            "normal": 123,
            "list": [
                1,
                2,
                3,
                lambda x: x,
                [11, 22, 33, lambda x: x, (44, 55, 66, lambda x: x)],
            ],  # noqa
            "tuple": (
                1,
                2,
                3,
                lambda x: x,
                (11, 22, 33, lambda x: x, [44, 55, 66, lambda x: x]),
            ),  # noqa
            "dict": {
                "subnormal": 456,
                "sublist": [4, 5, 6, lambda x: x],
                "subtuple": (4, 5, 6, lambda x: x),
                "subdict": {
                    "subsubnormal": 789,
                    "subsublist": [7, 8, 9, lambda x: x],
                    "subsubtuple": (7, 8, 9, lambda x: x),
                },
            },
        }

        test_compare = {
            "normal": 123,
            "list": [
                1,
                2,
                3,
                "funny return value",
                [11, 22, 33, "funny return value", (44, 55, 66, "funny return value")],
            ],  # noqa
            "tuple": (
                1,
                2,
                3,
                "funny return value",
                (11, 22, 33, "funny return value", [44, 55, 66, "funny return value"]),
            ),  # noqa
            "dict": {
                "subnormal": 456,
                "sublist": [4, 5, 6, "funny return value"],
                "subtuple": (4, 5, 6, "funny return value"),
                "subdict": {
                    "subsubnormal": 789,
                    "subsublist": [7, 8, 9, "funny return value"],
                    "subsubtuple": (7, 8, 9, "funny return value"),
                },
            },
        }

        test_out = call_callables(
            test_in,
            "funny return value",
        )

        self.assertEqual(test_out, test_compare)


class TestUtils(unittest.TestCase):
    def test_is_absolute(self):
        from plone.app.z3cform.utils import is_absolute

        self.assertTrue(is_absolute("https://plone.org/"))
        self.assertTrue(is_absolute("http://plone.org/"))
        self.assertTrue(is_absolute("webdav://plone.org/"))
        self.assertTrue(not is_absolute("./path/to/site"))
        self.assertTrue(not is_absolute("/resolveuid/"))

    def test_is_same_domain(self):
        from plone.app.z3cform.utils import is_same_domain

        # Those use the same protocol and are on the same domaain
        self.assertTrue(
            is_same_domain(
                "https://plone.org/doc1",
                "https://plone.org/doc2/doc3",
            )
        )

        # These are two completely different URLs
        self.assertTrue(
            not is_same_domain(
                "https://domain1.com",
                "https://anotherdomain.com",
            )
        )

        # Here, different transport protocols are used. Returning False.
        self.assertTrue(
            not is_same_domain(
                "https://plone.org",
                "http://plone.org",
            )
        )

    def test_unicode_control_character_removal(self):
        from plone.app.z3cform.utils import remove_invalid_xml_characters

        for x in range(32):
            if x in (9, 10, 13):
                self.assertTrue(remove_invalid_xml_characters(chr(x)) == chr(x))
            else:
                self.assertTrue(remove_invalid_xml_characters(chr(x)) == "")



from plone.app.z3cform.converters import DatetimeWidgetConverter
from plone.app.z3cform.converters import DateWidgetConverter
from plone.app.z3cform.interfaces import IDatetimeWidget
from plone.app.z3cform.interfaces import IDateWidget
from plone.app.z3cform.interfaces import ITimeWidget
from plone.app.z3cform.utils import dict_merge
from plone.app.z3cform.widgets.base import BaseWidget
from plone.app.z3cform.widgets.patterns import InputWidget
from plone.base import PloneMessageFactory as _
from z3c.form.browser.text import TextWidget
from z3c.form.interfaces import IFieldWidget
from z3c.form.widget import FieldWidget
from zope.i18n import translate
from zope.interface import implementer
from zope.interface import implementer_only


def get_date_options(request):
    calendar = request.locale.dates.calendars["gregorian"]
    return {
        "behavior": "native",
        "week-numbers": "show",
        "first-day": calendar.week.get("firstDay") == 1 and 1 or 0,
        "today": translate(_("Today"), context=request),
        "clear": translate(_("Clear"), context=request),
    }


@implementer_only(IDateWidget)
class DateWidget(BaseWidget, TextWidget):
    """Date widget for z3c.form.

    :param default_timezone: A Olson DB/pytz timezone identifier or a callback
                             returning such an identifier.
    :type default_timezone: String or callback

    :param default_time: Time used by converter as fallback if no time was set in UI.
    :type default_time: String or callback

    The default_timezone and default_time arguments are only used if a datewidget is
    used on a datetime field. If used on a date field they are ignored.
    """

    _base_type = "date"
    _converter = DateWidgetConverter
    _formater = "date"
    _formater_length = "short"

    default_timezone = None
    default_time = "00:00:00"

    pattern = "date-picker"
    pattern_options = BaseWidget.pattern_options.copy()

    def _base(self, **kw):
        return InputWidget(
            type=self._base_type,
            **kw,
        )

    def _base_args(self):
        """Method which will calculate _base class arguments.

        Returns (as python dictionary):
            - `pattern`: pattern name
            - `pattern_options`: pattern options
            - `name`: field name
            - `value`: field value

        :returns: Arguments which will be passed to _base
        :rtype: dict
        """
        args = super()._base_args()
        args["name"] = self.name
        args["value"] = (self.request.get(self.name, self.value) or "").strip()

        args.setdefault("pattern_options", {})
        if self.field.required:
            # Required fields should not have a "Clear" button
            args["pattern_options"]["clear"] = False
        args["pattern_options"] = dict_merge(
            get_date_options(self.request), args["pattern_options"]
        )

        return args

    def render(self):
        """Render widget.

        :returns: Widget's HTML.
        :rtype: string
        """
        if self.mode != "display":
            self.addClass("form-control")
            return super().render()

        if not self.value:
            return ""

        field_value = self._converter(self.field, self).toFieldValue(self.value)
        if field_value is self.field.missing_value:
            return ""

        formatter = self.request.locale.dates.getFormatter(
            self._formater,
            self._formater_length,
        )
        return formatter.format(field_value)


@implementer(IFieldWidget)
def DateFieldWidget(field, request):
    return FieldWidget(field, DateWidget(request))


@implementer_only(IDatetimeWidget)
class DatetimeWidget(DateWidget):
    """Datetime widget for z3c.form.

    :param default_timezone: A Olson DB/pytz timezone identifier or a callback
                             returning such an identifier.
    :type default_timezone: String or callback

    :param default_time: Time used by converter as fallback if no time was set in UI.
    :type default_time: String or callback
    """

    _base_type = "datetime-local"
    _converter = DatetimeWidgetConverter
    _formater = "dateTime"

    default_timezone = None
    default_time = "00:00:00"

    pattern = "datetime-picker"


@implementer(IFieldWidget)
def DatetimeFieldWidget(field, request):
    return FieldWidget(field, DatetimeWidget(request))


@implementer_only(ITimeWidget)
class TimeWidget(BaseWidget, TextWidget):
    pattern = ""

    def _base(self, **kw):
        return InputWidget(
            type="time",
            name=self.name,
            value=(self.request.get(self.name, self.value) or "").strip(),
            **kw,
        )

    def render(self):
        if self.mode != "display":
            self.addClass("form-control")
        return super().render()


@implementer(IFieldWidget)
def TimeFieldWidget(field, request):
    return FieldWidget(field, TimeWidget(request))



from OFS.interfaces import ISimpleItem
from plone.app.z3cform.interfaces import IAjaxSelectWidget
from plone.app.z3cform.interfaces import ISelectWidget
from plone.app.z3cform.utils import dict_merge
from plone.app.z3cform.utils import get_context_url
from plone.app.z3cform.utils import get_widget_form
from plone.app.z3cform.widgets.base import BaseWidget
from plone.app.z3cform.widgets.patterns import InputWidget
from plone.app.z3cform.widgets.patterns import SelectWidget
from plone.registry.interfaces import IRegistry
from Products.CMFCore.utils import getToolByName
from z3c.form import interfaces as form_ifaces
from z3c.form.browser.select import SelectWidget as z3cform_SelectWidget
from z3c.form.browser.text import TextWidget as z3cform_TextWidget
from z3c.form.interfaces import IEditForm
from z3c.form.interfaces import IFieldWidget
from z3c.form.interfaces import IForm
from z3c.form.interfaces import NO_VALUE
from z3c.form.widget import FieldWidget
from zope.component import getUtility
from zope.component import queryUtility
from zope.i18n import translate
from zope.interface import implementer
from zope.interface import implementer_only
from zope.schema import interfaces as schema_ifaces
from zope.schema.interfaces import IChoice
from zope.schema.interfaces import ICollection
from zope.schema.interfaces import ISequence
from zope.schema.interfaces import IVocabularyFactory

import collections


@implementer_only(ISelectWidget)
class SelectWidget(BaseWidget, z3cform_SelectWidget):
    """Select widget for z3c.form."""

    _base = SelectWidget

    pattern = "select2"
    pattern_options = BaseWidget.pattern_options.copy()

    separator = ";"
    noValueToken = ""
    noValueMessage = ""
    multiple = None
    orderable = False
    required = True

    @property
    def items(self):
        """
        Optionally handle ITreeVocabulary vocabs as dicts.
        """
        terms = self.terms
        if form_ifaces.ITerms.providedBy(terms):
            terms = terms.terms

        if schema_ifaces.ITreeVocabulary.providedBy(terms):
            groups = collections.OrderedDict()
            for group_term, option_terms in terms.items():
                group_widget = type(self)(self.request)
                group_widget.terms = option_terms
                group_label = group_term.title or group_term.value or group_term.token
                groups[group_label] = super(SelectWidget, group_widget).items
            return groups
        else:
            return super().items

    def _base_args(self):
        """Method which will calculate _base class arguments.

        Returns (as python dictionary):
            - `pattern`: pattern name
            - `pattern_options`: pattern options
            - `name`: field name
            - `value`: field value
            - `multiple`: field multiple
            - `items`: field items from which we can select to

        :returns: Arguments which will be passed to _base
        :rtype: dict
        """
        args = super()._base_args()
        args["name"] = self.name
        args["value"] = self.value
        args["multiple"] = self.multiple

        self.required = self.field.required

        options = args.setdefault("pattern_options", {})
        if self.multiple or ICollection.providedBy(self.field):
            args["multiple"] = self.multiple = True

        # ISequence represents an orderable collection
        if ISequence.providedBy(self.field) or self.orderable:
            options["orderable"] = True

        if self.multiple:
            options["separator"] = self.separator

        # Allow to clear field value if it is not required
        if not self.required:
            options["allowClear"] = True

        base_items = self.items
        if callable(base_items):
            # items used to be a property in all widgets, then in the select
            # widget it became a method, then in a few others too, but never in
            # all, so this was reverted to let it be a property again.  Let's
            # support both here to avoid breaking on some z3c.form versions.
            # See https://github.com/zopefoundation/z3c.form/issues/44
            base_items = base_items()

        def makeItem(item):
            """
            Gather the information needed by the widget for the given term.
            """
            if not isinstance(item["content"], str):
                item["content"] = translate(
                    item["content"], context=self.request, default=item["value"]
                )
            return (item["value"], item["content"])

        if isinstance(base_items, dict):
            items = collections.OrderedDict(
                (group_label, [makeItem(base_item) for base_item in group_options])
                for group_label, group_options in base_items.items()
            )
        else:
            items = [makeItem(item) for item in base_items]
        args["items"] = items

        return args

    def extract(self, default=NO_VALUE):
        """Override extract to handle delimited response values.
        Skip the vocabulary validation provided in the parent
        method, since it's not ever done for single selects."""
        if (
            self.name not in self.request
            and self.name + "-empty-marker" in self.request
        ):
            return []
        return self.request.get(self.name, default)


@implementer_only(IAjaxSelectWidget)
class AjaxSelectWidget(BaseWidget, z3cform_TextWidget):
    """Ajax select widget for z3c.form."""

    _base = InputWidget

    pattern = "select2"
    pattern_options = BaseWidget.pattern_options.copy()

    separator = ";"
    vocabulary = None
    vocabulary_view = "@@getVocabulary"
    orderable = False

    def _view_context(self):
        view_context = get_widget_form(self)
        # For EditForms and non-Forms (in tests), the vocabulary is looked
        # up on the context, otherwise on the view
        if IEditForm.providedBy(view_context):
            if self.is_subform_widget():
                view_context = self.form.parentForm.context
            elif not ISimpleItem.providedBy(self.context):
                view_context = self.form.context
            else:
                view_context = self.context
        elif not IForm.providedBy(view_context):
            view_context = self.context
        return view_context

    def get_vocabulary(self):
        if self.vocabulary and isinstance(self.vocabulary, str):
            factory = queryUtility(
                IVocabularyFactory,
                self.vocabulary,
            )
            if factory:
                return factory(self._view_context())
        return self.vocabulary

    def display_items(self):
        if self.value:
            tokens = self.value.split(self.separator)
            vocabulary = self.get_vocabulary()
            for token in tokens:
                item = {"token": token, "title": token}
                if vocabulary is not None:
                    try:
                        item["title"] = vocabulary.getTermByToken(token).title
                    except LookupError:
                        pass
                yield item

    def has_multiple_values(self):
        return self.value and self.value.split(self.separator)

    def _ajaxselect_options(self):
        options = {
            "separator": self.separator,
        }
        if self.vocabulary:
            options["vocabularyUrl"] = "{}/{}?name={}".format(
                get_context_url(self._view_context()),
                self.vocabulary_view,
                self.vocabulary,
            )
            field_name = self.field and self.field.__name__ or None
            if field_name:
                options["vocabularyUrl"] += f"&field={field_name}"
            vocabulary = self.get_vocabulary()
            if vocabulary is not None and self.value:
                options["initialValues"] = dict()
                for token in self.value.split(self.separator):
                    try:
                        term = vocabulary.getTermByToken(token)
                        options["initialValues"][term.token] = term.title
                    except LookupError:
                        options["initialValues"][token] = token

        return options

    def update(self):
        super().update()
        field = getattr(self, "field", None)
        field = getattr(field, "value_type", field)
        if (
            not self.vocabulary
            and field is not None
            and getattr(field, "vocabularyName", None)
        ):
            self.vocabulary = field.vocabularyName

    def _base_args(self):
        """Method which will calculate _base class arguments.

        Returns (as python dictionary):
            - `pattern`: pattern name
            - `pattern_options`: pattern options
            - `name`: field name
            - `value`: field value

        :returns: Arguments which will be passed to _base
        :rtype: dict
        """
        args = super()._base_args()
        args["name"] = self.name
        args["value"] = self.value
        args.setdefault("pattern_options", {})
        context = self.context
        field = None

        if IChoice.providedBy(self.field):
            args["pattern_options"]["maximumSelectionSize"] = 1
            field = self.field
        elif ICollection.providedBy(self.field):
            field = self.field.value_type
        if IChoice.providedBy(field):
            args["pattern_options"]["allowNewItems"] = "false"

        args["pattern_options"] = dict_merge(
            self._ajaxselect_options(), args["pattern_options"]
        )

        if field and getattr(field, "vocabulary", None):
            form_url = self.request.getURL()
            source_url = "{:s}/++widget++{:s}/@@getSource".format(
                form_url,
                self.name,
            )
            args["pattern_options"]["vocabularyUrl"] = source_url

        # ISequence represents an orderable collection
        if ISequence.providedBy(self.field) or self.orderable:
            args["pattern_options"]["orderable"] = True

        if self.vocabulary == "plone.app.vocabularies.Keywords":
            membership = getToolByName(context, "portal_membership")
            user = membership.getAuthenticatedMember()

            registry = getUtility(IRegistry)
            roles_allowed_to_add_keywords = registry.get(
                "plone.roles_allowed_to_add_keywords", set()
            )
            roles = set(user.getRolesInContext(context))
            allowNewItems = bool(
                roles.intersection(roles_allowed_to_add_keywords),
            )
            args["pattern_options"]["allowNewItems"] = str(
                allowNewItems,
            ).lower()

        return args


@implementer(IFieldWidget)
def SelectFieldWidget(field, request):
    return FieldWidget(field, SelectWidget(request))


@implementer(IFieldWidget)
def AjaxSelectFieldWidget(field, request, extra=None):
    if extra is not None:
        request = extra
    return FieldWidget(field, AjaxSelectWidget(request))



from datetime import date
from datetime import datetime
from datetime import time
from json import loads
from lxml import html
from plone.app.contentlisting.contentlisting import ContentListing
from plone.app.testing import setRoles
from plone.app.testing import TEST_USER_ID
from plone.app.z3cform.tests.layer import PAZ3CForm_INTEGRATION_TESTING
from plone.app.z3cform.widgets.base import BaseWidget
from plone.app.z3cform.widgets.base import PatternNotImplemented
from plone.app.z3cform.widgets.datetime import DateWidget
from plone.app.z3cform.widgets.relateditems import RelatedItemsWidget
from plone.autoform.directives import widget
from plone.autoform.form import AutoExtensibleForm
from plone.base.interfaces import IMarkupSchema
from plone.dexterity.fti import DexterityFTI
from plone.registry.interfaces import IRegistry
from plone.supermodel.model import Schema
from plone.testing.zca import UNIT_TESTING
from plone.uuid.interfaces import IUUID
from unittest import mock
from unittest.mock import Mock
from z3c.form.form import EditForm
from z3c.form.form import Form
from z3c.form.interfaces import IFormLayer
from z3c.form.widget import FieldWidget
from z3c.relationfield.relation import RelationValue
from z3c.relationfield.schema import RelationChoice
from z3c.relationfield.schema import RelationList
from zope.component import getMultiAdapter
from zope.component import getUtility
from zope.component import provideUtility
from zope.component.globalregistry import base
from zope.globalrequest import setRequest
from zope.interface import alsoProvides
from zope.interface import Interface
from zope.interface import provider
from zope.intid.interfaces import IIntIds
from zope.pagetemplate.interfaces import IPageTemplate
from zope.publisher.browser import TestRequest
from zope.schema import BytesLine
from zope.schema import Choice
from zope.schema import Date
from zope.schema import Datetime
from zope.schema import List
from zope.schema import Set
from zope.schema import TextLine
from zope.schema import Time
from zope.schema import Tuple
from zope.schema import vocabulary
from zope.schema.interfaces import IVocabularyFactory
from zope.schema.vocabulary import SimpleTerm
from zope.schema.vocabulary import SimpleVocabulary

import json
import pytz
import unittest


@provider(IVocabularyFactory)
def example_vocabulary_factory(context, query=None):
    items = ["One", "Two", "Three"]
    tmp = SimpleVocabulary(
        [
            SimpleTerm(
                item.lower(),  # value
                token=f"token_{item.lower()}",
                title=item,
            )
            for item in items
            if query is None or query.lower() in item.lower()
        ],
    )
    tmp.test = 1
    return tmp


class BaseWidgetTests(unittest.TestCase):
    def setUp(self):
        self.request = TestRequest(environ={"HTTP_ACCEPT_LANGUAGE": "en"})
        self.field = TextLine(__name__="textlinefield")
        self.maxDiff = 999999

    def test_widget_pattern_notimplemented(self):
        widget = BaseWidget(self.request)
        widget.field = self.field

        self.assertRaises(
            PatternNotImplemented,
            widget._base_args,
        )

        widget.pattern = "example"

        self.assertEqual(
            {
                "pattern": "example",
                "pattern_options": {},
            },
            widget._base_args(),
        )

    def test_widget_base_notimplemented(self):
        from plone.app.z3cform.widgets.base import BaseWidget
        from plone.app.z3cform.widgets.patterns import InputWidget

        widget = BaseWidget(self.request)
        widget.field = self.field
        widget.pattern = "example"

        self.assertRaises(
            PatternNotImplemented,
            widget.render,
        )

        widget._base = InputWidget

        self.assertEqual(
            '<input class="pat-example" type="text"/>',
            widget.render(),
        )

    def test_widget_base_custom_css(self):
        from plone.app.z3cform.widgets.base import BaseWidget
        from plone.app.z3cform.widgets.patterns import InputWidget

        widget = BaseWidget(self.request)
        widget.field = self.field
        widget.pattern = "example"
        widget.klass = "very-custom-class"
        widget._base = InputWidget

        self.assertEqual(
            '<input class="pat-example very-custom-class" type="text"/>',
            widget.render(),
        )

    def test_widget_base_pattern_options_with_functions(self):
        from plone.app.z3cform.widgets.base import BaseWidget
        from plone.app.z3cform.widgets.patterns import InputWidget

        widget = BaseWidget(self.request)
        widget.context = "testcontext"
        widget.field = self.field
        widget.pattern = "example"
        widget._base = InputWidget
        widget.pattern_options = {
            "subdict": {
                "subsubnormal": 789,
                "subsublist": [7, 8, 9, lambda x: x],
                "subsubtuple": (7, 8, 9, lambda x: x),
            },
        }
        output = widget.render()
        # output is something like
        #
        # <input class="pat-example"
        #        type="text"
        #        data-pat-example="$JSON_ENCODED_OPTIONS" />'
        self.assertRegex(widget.render(), "<input .*/>")
        # We cannot foresee how the options are encoded
        # so we will extract the attributes with lxml
        # and be sure that they will match what we expect
        observed_attrib = html.fromstring(output).attrib
        self.assertEqual(
            sorted(observed_attrib),
            ["class", "data-pat-example", "type"],
        )
        self.assertEqual(observed_attrib["class"], "pat-example")
        self.assertEqual(observed_attrib["type"], "text")
        self.assertDictEqual(
            loads(observed_attrib["data-pat-example"]),
            {
                "subdict": {
                    "subsubnormal": 789,
                    "subsublist": [7, 8, 9, "testcontext"],
                    "subsubtuple": [7, 8, 9, "testcontext"],
                },
            },
        )


class DateWidgetTests(unittest.TestCase):
    def setUp(self):
        from plone.app.z3cform.widgets.datetime import DateWidget

        self.request = TestRequest(environ={"HTTP_ACCEPT_LANGUAGE": "en"})
        self.field = Date(__name__="datefield")
        self.field.required = False
        self.widget = DateWidget(self.request)
        self.widget.field = self.field
        self.widget.pattern_options = {"date": {"firstDay": 0}}

    def test_widget(self):
        self.assertEqual(
            {
                "name": None,
                "pattern": "date-picker",
                "pattern_options": {
                    "behavior": "native",
                    "clear": "Clear",
                    "date": {"firstDay": 0},
                    "first-day": 0,
                    "today": "Today",
                    "week-numbers": "show",
                },
                "value": "",
            },
            self.widget._base_args(),
        )

    def test_widget_required(self):
        """Required fields should not have a "Clear" button."""
        self.field.required = True
        base_args = self.widget._base_args()
        self.assertEqual(base_args["pattern_options"]["clear"], False)

    def test_data_converter(self):
        from plone.app.z3cform.widgets.datetime import DateWidgetConverter

        converter = DateWidgetConverter(self.field, self.widget)

        self.assertEqual(
            converter.field.missing_value,
            converter.toFieldValue(""),
        )

        self.assertEqual(
            date(2000, 10, 30),
            converter.toFieldValue("2000-10-30"),
        )

        self.assertEqual(
            date(21, 10, 30),
            converter.toFieldValue("21-10-30"),
        )

        self.assertEqual(
            "",
            converter.toWidgetValue(converter.field.missing_value),
        )

        self.assertEqual(
            "2000-10-30",
            converter.toWidgetValue(date(2000, 10, 30)),
        )

        self.assertEqual(
            "21-10-30",
            converter.toWidgetValue(date(21, 10, 30)),
        )

    def test_fieldwidget(self):
        from plone.app.z3cform.widgets.datetime import DateFieldWidget
        from plone.app.z3cform.widgets.datetime import DateWidget

        field = Mock(__name__="field", title="", required=True)
        request = Mock()
        widget = DateFieldWidget(field, request)
        self.assertTrue(isinstance(widget, DateWidget))
        self.assertIs(widget.field, field)
        self.assertIs(widget.request, request)

    def test_dateformatter(self):
        self.widget.value = "2022-08-17"
        self.assertIn(' value="2022-08-17" ', self.widget.render())

        self.widget.mode = "display"
        self.assertEqual("8/17/22", self.widget.render())

        self.widget._formater_length = "medium"
        self.assertEqual("Aug 17, 2022", self.widget.render())

        self.widget._formater_length = "long"
        self.assertEqual("August 17, 2022", self.widget.render())

        self.widget._formater_length = "full"
        self.assertEqual("Wednesday, August 17, 2022", self.widget.render())

        # unknown formatter length
        self.widget._formater_length = "foo"
        with self.assertRaises(ValueError):
            self.widget.render()


class DatetimeWidgetTests(unittest.TestCase):
    def setUp(self):
        from plone.app.z3cform.widgets.datetime import DatetimeWidget

        self.request = TestRequest(environ={"HTTP_ACCEPT_LANGUAGE": "en"})
        self.field = Datetime(__name__="datetimefield")
        self.field.required = False
        self.widget = DatetimeWidget(self.request)
        self.widget.field = self.field
        self.widget.pattern_options = {
            "date": {"firstDay": 0},
            "time": {"interval": 15},
        }

    def test_widget(self):
        self.assertEqual(
            {
                "name": None,
                "pattern": "datetime-picker",
                "pattern_options": {
                    "behavior": "native",
                    "clear": "Clear",
                    "date": {"firstDay": 0},
                    "first-day": 0,
                    "time": {"interval": 15},
                    "today": "Today",
                    "week-numbers": "show",
                },
                "value": "",
            },
            self.widget._base_args(),
        )

    def test_widget_required(self):
        """Required fields should not have a "Clear" button."""
        self.field.required = True
        base_args = self.widget._base_args()
        self.assertEqual(base_args["pattern_options"]["clear"], False)

    def test_data_converter(self):
        from plone.app.z3cform.widgets.datetime import DatetimeWidgetConverter

        converter = DatetimeWidgetConverter(self.field, self.widget)

        self.assertEqual(
            converter.toFieldValue(""),
            converter.field.missing_value,
        )

        self.assertEqual(
            converter.toFieldValue("2000-10-30T15:40"),
            datetime(2000, 10, 30, 15, 40),
        )

        self.assertEqual(
            converter.toFieldValue("21-10-30T15:40"),
            datetime(21, 10, 30, 15, 40),
        )

        self.assertEqual(
            converter.toWidgetValue(converter.field.missing_value),
            "",
        )

        self.assertEqual(
            converter.toWidgetValue(datetime(2000, 10, 30, 15, 40)),
            "2000-10-30T15:40",
        )

        self.assertEqual(
            converter.toWidgetValue(datetime(21, 10, 30, 15, 40)),
            "21-10-30T15:40",
        )

    def test_data_converter__no_timezone(self):
        """When no timezone is set, don't apply one."""
        from plone.app.z3cform.widgets.datetime import DatetimeWidgetConverter

        context = Mock()

        dt = datetime(2013, 11, 13, 10, 20)
        setattr(context, self.field.getName(), dt)
        self.widget.context = context
        self.widget.default_timezone = None

        converter = DatetimeWidgetConverter(self.field, self.widget)
        self.assertEqual(
            converter.toFieldValue("2013-11-13T10:20"),
            datetime(2013, 11, 13, 10, 20),
        )

        # cleanup
        self.widget.context = None
        self.widget.default_timezone = None

    def test_data_converter__timezone_id(self):
        """When a (pytz) timezone id is set, use that."""
        from plone.app.z3cform.widgets.datetime import DatetimeWidgetConverter

        context = Mock()

        dt = datetime(2013, 11, 13, 10, 20)
        setattr(context, self.field.getName(), dt)
        self.widget.context = context
        self.widget.default_timezone = "Europe/Amsterdam"
        tz = pytz.timezone("Europe/Amsterdam")

        converter = DatetimeWidgetConverter(self.field, self.widget)
        self.assertEqual(
            converter.toFieldValue("2013-11-13T10:20"),
            tz.localize(datetime(2013, 11, 13, 10, 20)),
        )

        # cleanup
        self.widget.context = None
        self.widget.default_timezone = None

    def test_data_converter__timezone_callback(self):
        """When a timezone callback is set, returning a (pytz) timezone id,
        use that.
        """
        from plone.app.z3cform.widgets.datetime import DatetimeWidgetConverter

        context = Mock()

        dt = datetime(2013, 11, 13, 10, 20)
        setattr(context, self.field.getName(), dt)
        self.widget.context = context
        self.widget.default_timezone = lambda context: "Europe/Amsterdam"
        tz = pytz.timezone("Europe/Amsterdam")

        converter = DatetimeWidgetConverter(self.field, self.widget)
        self.assertEqual(
            converter.toFieldValue("2013-11-13T10:20"),
            tz.localize(datetime(2013, 11, 13, 10, 20)),
        )

        # cleanup
        self.widget.context = None
        self.widget.default_timezone = None

    def test_fieldwidget(self):
        from plone.app.z3cform.widgets.datetime import DatetimeFieldWidget
        from plone.app.z3cform.widgets.datetime import DatetimeWidget

        field = Mock(__name__="field", title="", required=True)
        request = Mock()
        widget = DatetimeFieldWidget(field, request)
        self.assertTrue(isinstance(widget, DatetimeWidget))
        self.assertIs(widget.field, field)
        self.assertIs(widget.request, request)

    def test_datetimeformatter(self):
        self.widget.value = "2022-08-17T12:00"
        self.assertIn(' value="2022-08-17T12:00" ', self.widget.render())

        self.widget.mode = "display"
        self.assertEqual("8/17/22 12:00 PM", self.widget.render())

        self.widget._formater_length = "medium"
        self.assertEqual("Aug 17, 2022 12:00:00 PM", self.widget.render())

        self.widget._formater_length = "long"
        self.assertEqual("August 17, 2022 12:00:00 PM +000", self.widget.render())

        self.widget._formater_length = "full"
        self.assertEqual(
            "Wednesday, August 17, 2022 12:00:00 PM +000", self.widget.render()
        )

        # unknown formatter length
        self.widget._formater_length = "foo"
        with self.assertRaises(ValueError):
            self.widget.render()


class TimeWidgetTests(unittest.TestCase):
    def setUp(self):
        from plone.app.z3cform.widgets.datetime import TimeWidget

        self.request = TestRequest(environ={"HTTP_ACCEPT_LANGUAGE": "en"})
        self.field = Time(__name__="timefield")
        self.field.required = False
        self.widget = TimeWidget(self.request)
        self.widget.field = self.field

    def test_widget(self):
        self.assertIn('<input type="time"', self.widget.render())

    def test_data_converter(self):
        from plone.app.z3cform.converters import TimeWidgetConverter

        converter = TimeWidgetConverter(self.field, self.widget)

        self.assertEqual(
            converter.toFieldValue(""),
            converter.field.missing_value,
        )

        self.assertEqual(
            converter.toFieldValue("15:40"),
            time(15, 40),
        )

        self.assertEqual(
            converter.toWidgetValue(converter.field.missing_value),
            "",
        )

        self.assertEqual(
            converter.toWidgetValue(time(15, 40)),
            "15:40",
        )

    def test_fieldwidget(self):
        from plone.app.z3cform.widgets.datetime import TimeFieldWidget
        from plone.app.z3cform.widgets.datetime import TimeWidget

        field = Mock(__name__="field", title="", required=True)
        request = Mock()
        widget = TimeFieldWidget(field, request)
        self.assertTrue(isinstance(widget, TimeWidget))
        self.assertIs(widget.field, field)
        self.assertIs(widget.request, request)


class SelectWidgetTests(unittest.TestCase):
    def setUp(self):
        self.request = TestRequest(environ={"HTTP_ACCEPT_LANGUAGE": "en"})
        alsoProvides(self.request, IFormLayer)

        # ITerms Adapters are needed for data converter
        from z3c.form import term

        import zope.component

        zope.component.provideAdapter(term.CollectionTerms)
        zope.component.provideAdapter(term.CollectionTermsVocabulary)
        zope.component.provideAdapter(term.CollectionTermsSource)

    def tearDown(self):
        from z3c.form import term

        base.unregisterAdapter(term.CollectionTerms)
        base.unregisterAdapter(term.CollectionTermsVocabulary)
        base.unregisterAdapter(term.CollectionTermsSource)

    def test_widget(self):
        from plone.app.z3cform.widgets.select import SelectWidget

        widget = SelectWidget(self.request)
        widget.id = "test-widget"
        widget.field = Choice(
            __name__="selectfield",
            values=["one", "two", "three"],
        )
        widget.terms = widget.field.vocabulary
        widget.field.required = True
        self.assertEqual(
            {
                "multiple": None,
                "name": None,
                "pattern_options": {},
                "pattern": "select2",
                "value": (),
                "items": [
                    ("one", "one"),
                    ("two", "two"),
                    ("three", "three"),
                ],
            },
            widget._base_args(),
        )

        widget.field.required = False
        self.assertEqual(
            {
                "multiple": None,
                "name": None,
                "pattern_options": {"allowClear": True},
                "pattern": "select2",
                "value": (),
                "items": [
                    ("", ""),
                    ("one", "one"),
                    ("two", "two"),
                    ("three", "three"),
                ],
            },
            widget._base_args(),
        )

        widget.field.required = True
        widget.multiple = True
        self.assertEqual(
            {
                "multiple": True,
                "name": None,
                "pattern_options": {"separator": ";"},
                "pattern": "select2",
                "value": (),
                "items": [
                    ("one", "one"),
                    ("two", "two"),
                    ("three", "three"),
                ],
            },
            widget._base_args(),
        )

        widget.field.required = False
        widget.multiple = True
        self.assertEqual(
            {
                "multiple": True,
                "name": None,
                "pattern_options": {"allowClear": True, "separator": ";"},
                "pattern": "select2",
                "value": (),
                "items": [
                    ("one", "one"),
                    ("two", "two"),
                    ("three", "three"),
                ],
            },
            widget._base_args(),
        )

        widget.value = "one"
        self.assertEqual(
            {
                "multiple": True,
                "name": None,
                "pattern_options": {"allowClear": True, "separator": ";"},
                "pattern": "select2",
                "value": ("one"),
                "items": [
                    ("one", "one"),
                    ("two", "two"),
                    ("three", "three"),
                ],
            },
            widget._base_args(),
        )

    def test_widget_list_orderable(self):
        from plone.app.z3cform.widgets.select import SelectWidget

        widget = SelectWidget(self.request)
        widget.id = "test-widget"
        widget.separator = "."
        widget.field = List(
            __name__="selectfield",
            value_type=Choice(values=["one", "two", "three"]),
        )
        widget.terms = widget.field.value_type.vocabulary
        self.assertEqual(
            {
                "multiple": True,
                "name": None,
                "pattern_options": {"orderable": True, "separator": "."},
                "pattern": "select2",
                "value": (),
                "items": [
                    ("one", "one"),
                    ("two", "two"),
                    ("three", "three"),
                ],
            },
            widget._base_args(),
        )

    def test_widget_tuple_orderable(self):
        from plone.app.z3cform.widgets.select import SelectWidget

        widget = SelectWidget(self.request)
        widget.id = "test-widget"
        widget.field = Tuple(
            __name__="selectfield",
            value_type=Choice(values=["one", "two", "three"]),
        )
        widget.terms = widget.field.value_type.vocabulary
        self.assertEqual(
            {
                "multiple": True,
                "name": None,
                "pattern_options": {"orderable": True, "separator": ";"},
                "pattern": "select2",
                "value": (),
                "items": [
                    ("one", "one"),
                    ("two", "two"),
                    ("three", "three"),
                ],
            },
            widget._base_args(),
        )

    def test_widget_set_not_orderable(self):
        from plone.app.z3cform.widgets.select import SelectWidget

        widget = SelectWidget(self.request)
        widget.id = "test-widget"
        # A set is not orderable
        widget.field = Set(
            __name__="selectfield",
            value_type=Choice(values=["one", "two", "three"]),
        )
        widget.terms = widget.field.value_type.vocabulary
        self.assertEqual(
            {
                "multiple": True,
                "name": None,
                "pattern_options": {"separator": ";"},
                "pattern": "select2",
                "value": (),
                "items": [
                    ("one", "one"),
                    ("two", "two"),
                    ("three", "three"),
                ],
            },
            widget._base_args(),
        )

    def test_widget_extract(self):
        from plone.app.z3cform.widgets.select import SelectWidget

        widget = SelectWidget(self.request)
        widget.field = Choice(
            __name__="selectfield",
            values=["one", "two", "three"],
        )
        widget.name = "selectfield"
        self.request.form["selectfield"] = "one"
        self.assertEqual(widget.extract(), "one")
        widget.multiple = True
        self.request.form["selectfield"] = "one;two"
        self.assertEqual(widget.extract(), "one;two")

    def test_data_converter_list(self):
        from plone.app.z3cform.converters import SelectWidgetConverter
        from plone.app.z3cform.widgets.select import SelectWidget

        field = List(
            __name__="listfield",
            value_type=Choice(
                __name__="selectfield",
                values=["one", "two", "three"],
            ),
        )
        widget = SelectWidget(self.request)
        widget.field = field
        widget.multiple = True
        converter = SelectWidgetConverter(field, widget)

        self.assertEqual(
            converter.toFieldValue(""),
            field.missing_value,
        )

        self.assertEqual(
            converter.toFieldValue("one;two;three"),
            ["one", "two", "three"],
        )

        self.assertEqual(
            converter.toWidgetValue([]),
            [],
        )

        widget.separator = ","
        self.assertEqual(
            converter.toFieldValue("one,two,three"),
            ["one", "two", "three"],
        )
        self.assertRaises(
            LookupError,
            converter.toFieldValue,
            "one;two;three",
        )

        self.assertEqual(
            converter.toWidgetValue(["one", "two", "three"]),
            ["one", "two", "three"],
        )

    def test_data_converter_tuple(self):
        from plone.app.z3cform.converters import SelectWidgetConverter
        from plone.app.z3cform.widgets.select import SelectWidget

        field = Tuple(
            __name__="tuplefield",
            value_type=Choice(
                __name__="selectfield",
                values=["one", "two", "three"],
            ),
        )
        widget = SelectWidget(self.request)
        widget.field = field
        widget.multiple = True
        converter = SelectWidgetConverter(field, widget)

        self.assertEqual(
            converter.toFieldValue(""),
            field.missing_value,
        )

        self.assertEqual(
            converter.toFieldValue("one;two;three"),
            ("one", "two", "three"),
        )

        self.assertEqual(
            converter.toWidgetValue(tuple()),
            [],
        )

        self.assertEqual(
            converter.toWidgetValue(("one", "two", "three")),
            ["one", "two", "three"],
        )

    def test_data_converter_handles_empty_value(self):
        from plone.app.z3cform.converters import SelectWidgetConverter
        from plone.app.z3cform.widgets.select import SelectWidget

        field = Tuple(
            __name__="tuplefield",
            value_type=Choice(__name__="selectfield", values=["one", "two", "three"]),
        )
        widget = SelectWidget(self.request)
        widget.field = field
        widget.multiple = True
        converter = SelectWidgetConverter(field, widget)

        self.assertEqual(
            converter.toFieldValue(("",)),
            field.missing_value,
        )

    def test_widget_optgroup(self):
        """
        If the widget vocabulary is a mapping <optgroup>'s are rendered.
        """
        from plone.app.z3cform.widgets.select import SelectWidget
        from z3c.form import term

        widget = SelectWidget(self.request)
        widget.field = Choice(
            vocabulary=vocabulary.TreeVocabulary.fromDict(
                {
                    ("foo_group", "Foo Group"): {
                        ("bar_group", "Bar Group"): {},
                        ("qux_group", "Qux Group"): {},
                    },
                    ("corge_group", "Corge Group"): {
                        ("grault_group", "Grault Group"): {},
                        ("garply_group", "Garply Group"): {},
                    },
                }
            )
        )
        # Usse term.CollectionTermsVocabulary to simulate a named vocabulary
        # factory lookup
        widget.terms = term.CollectionTermsVocabulary(
            context=None,
            request=self.request,
            form=None,
            field=None,
            widget=widget,
            vocabulary=widget.field.vocabulary,
        )
        widget.updateTerms()
        html = widget.render()
        self.assertNotIn(
            '<option value="foo_group">',
            html,
            "Top level vocab item rendered as <option...>",
        )
        self.assertIn(
            '<optgroup label="Foo Group">',
            html,
            "Rendered select widget missing an <optgroup...>",
        )

        base_args = widget._base_args()
        pattern_widget = widget._base(**base_args)
        items = pattern_widget.items
        self.assertIsInstance(items, dict, "Wrong widget items type")


class AjaxSelectWidgetTests(unittest.TestCase):
    layer = UNIT_TESTING
    maxDiff = None

    def setUp(self):
        self.request = TestRequest(environ={"HTTP_ACCEPT_LANGUAGE": "en"})
        provideUtility(example_vocabulary_factory, name="example")

    def test_widget(self):
        from plone.app.z3cform.widgets.select import AjaxSelectWidget

        widget = AjaxSelectWidget(self.request)
        widget.update()
        self.assertEqual(
            {
                "name": None,
                "value": "",
                "pattern": "select2",
                "pattern_options": {"separator": ";"},
            },
            widget._base_args(),
        )

        widget.vocabulary = "example"
        self.assertEqual(
            widget._base_args(),
            {
                "name": None,
                "value": "",
                "pattern": "select2",
                "pattern_options": {
                    "vocabularyUrl": "/@@getVocabulary?name=example",
                    "separator": ";",
                },
            },
        )

        widget.value = "token_three;token_two"
        self.assertDictEqual(
            {
                "name": None,
                "value": "token_three;token_two",
                "pattern": "select2",
                "pattern_options": {
                    "vocabularyUrl": "/@@getVocabulary?name=example",
                    "initialValues": {
                        "token_three": "Three",
                        "token_two": "Two",
                    },
                    "separator": ";",
                },
            },
            widget._base_args(),
        )

    def test_widget_list_orderable(self):
        from plone.app.z3cform.widgets.select import AjaxSelectWidget

        widget = AjaxSelectWidget(self.request)
        widget.field = List(__name__="selectfield")
        self.assertEqual(
            {
                "name": None,
                "value": "",
                "pattern": "select2",
                "pattern_options": {"orderable": True, "separator": ";"},
            },
            widget._base_args(),
        )

    def test_widget_tuple_orderable(self):
        from plone.app.z3cform.widgets.select import AjaxSelectWidget

        widget = AjaxSelectWidget(self.request)
        widget.field = Tuple(__name__="selectfield")
        self.assertEqual(
            {
                "name": None,
                "value": "",
                "pattern": "select2",
                "pattern_options": {"orderable": True, "separator": ";"},
            },
            widget._base_args(),
        )

    def test_widget_set_not_orderable(self):
        from plone.app.z3cform.widgets.select import AjaxSelectWidget

        widget = AjaxSelectWidget(self.request)
        # A set is not orderable
        widget.field = Set(__name__="selectfield")
        self.assertEqual(
            {
                "name": None,
                "value": "",
                "pattern": "select2",
                "pattern_options": {"separator": ";"},
            },
            widget._base_args(),
        )

    def test_widget_choice(self):
        from plone.app.z3cform.widgets.select import AjaxSelectWidget
        from zope.schema.interfaces import ISource

        widget = AjaxSelectWidget(self.request)
        source = Mock()
        alsoProvides(source, ISource)
        widget.field = Choice(__name__="choicefield", source=source)
        widget.name = "choicefield"
        self.assertEqual(
            {
                "name": "choicefield",
                "value": "",
                "pattern": "select2",
                "pattern_options": {
                    "separator": ";",
                    "maximumSelectionSize": 1,
                    "allowNewItems": "false",
                    "vocabularyUrl": "http://127.0.0.1/++widget++choicefield/@@getSource",
                },
            },
            widget._base_args(),
        )

    def test_widget_addform_url_on_addform(self):
        from plone.app.z3cform.widgets.select import AjaxSelectWidget

        widget = AjaxSelectWidget(self.request)
        form = Mock(parentForm=None)
        from z3c.form.interfaces import IAddForm
        from zope.interface import directlyProvides  # noqa

        directlyProvides(form, IAddForm)  # noqa
        form.request = {"URL": "http://addform_url"}
        widget.form = form
        self.assertEqual(
            {
                "name": None,
                "value": "",
                "pattern": "select2",
                "pattern_options": {"separator": ";"},
            },
            widget._base_args(),
        )
        widget.vocabulary = "vocabulary1"
        self.assertEqual(
            {
                "name": None,
                "value": "",
                "pattern": "select2",
                "pattern_options": {
                    "separator": ";",
                    "vocabularyUrl": "http://addform_url/@@getVocabulary?name=vocabulary1",
                },
            },
            widget._base_args(),
        )

    def test_data_converter_list(self):
        from plone.app.z3cform.converters import AjaxSelectWidgetConverter
        from plone.app.z3cform.widgets.select import AjaxSelectWidget

        field = List(__name__="listfield", value_type=TextLine())
        widget = AjaxSelectWidget(self.request)
        widget.field = field
        converter = AjaxSelectWidgetConverter(field, widget)

        self.assertEqual(
            converter.toFieldValue(""),
            field.missing_value,
        )

        self.assertEqual(
            converter.toFieldValue("123;456;789"),
            ["123", "456", "789"],
        )

        self.assertEqual(
            converter.toWidgetValue([]),
            None,
        )

        self.assertEqual(
            converter.toWidgetValue(["123", "456", "789"]),
            "123;456;789",
        )

    def test_data_converter_collection_with_vocabulary(self):
        from plone.app.z3cform.converters import AjaxSelectWidgetConverter
        from plone.app.z3cform.widgets.select import AjaxSelectWidget

        field = Tuple(
            __name__="listfield",
            value_type=Choice(
                vocabulary="example",
            ),
        )
        widget = AjaxSelectWidget(self.request)
        widget.field = field
        converter = AjaxSelectWidgetConverter(field, widget)

        self.assertEqual(
            converter.toFieldValue(""),
            field.missing_value,
        )

        self.assertEqual(
            converter.toFieldValue("token_one;token_two;token_three"),
            ("one", "two", "three"),
        )

        self.assertEqual(
            converter.toWidgetValue([]),
            None,
        )

        self.assertEqual(
            converter.toWidgetValue(["123", "456", "789"]),
            "123;456;789",
        )

    def test_data_converter_tuple(self):
        from plone.app.z3cform.converters import AjaxSelectWidgetConverter
        from plone.app.z3cform.widgets.select import AjaxSelectWidget

        field = Tuple(__name__="tuplefield", value_type=TextLine())
        widget = AjaxSelectWidget(self.request)
        widget.field = field
        converter = AjaxSelectWidgetConverter(field, widget)

        self.assertEqual(
            converter.toFieldValue(""),
            field.missing_value,
        )

        self.assertEqual(
            converter.toFieldValue("123;456;789"),
            ("123", "456", "789"),
        )

        self.assertEqual(
            converter.toWidgetValue(tuple()),
            None,
        )

        self.assertEqual(
            converter.toWidgetValue(("123", "456", "789")),
            "123;456;789",
        )

    def test_fieldwidget(self):
        from plone.app.z3cform.widgets.select import AjaxSelectFieldWidget
        from plone.app.z3cform.widgets.select import AjaxSelectWidget

        field = Mock(__name__="field", title="", required=True)
        request = Mock()
        widget = AjaxSelectFieldWidget(field, request)
        self.assertTrue(isinstance(widget, AjaxSelectWidget))
        self.assertIs(widget.field, field)
        self.assertIs(widget.request, request)

    def test_fieldwidget_sequence(self):
        from plone.app.z3cform.widgets.select import AjaxSelectFieldWidget
        from plone.app.z3cform.widgets.select import AjaxSelectWidget

        field = Mock(__name__="field", title="", required=True)
        vocabulary = Mock()
        request = Mock()
        widget = AjaxSelectFieldWidget(field, vocabulary, request)
        self.assertTrue(isinstance(widget, AjaxSelectWidget))
        self.assertIs(widget.field, field)
        self.assertIs(widget.request, request)


class AjaxSelectWidgetIntegrationTests(unittest.TestCase):
    layer = PAZ3CForm_INTEGRATION_TESTING

    def setUp(self):
        self.request = TestRequest(environ={"HTTP_ACCEPT_LANGUAGE": "en"})

    def test_keywords_can_add(self):
        from plone.app.z3cform.widgets.select import AjaxSelectWidget

        portal = self.layer["portal"]
        setRoles(portal, TEST_USER_ID, ["Manager"])
        widget = AjaxSelectWidget(self.request)
        widget.context = portal
        widget.vocabulary = "plone.app.vocabularies.Keywords"
        self.assertEqual(
            widget._base_args()["pattern_options"]["allowNewItems"],
            "true",
        )

    def test_keywords_cannot_add(self):
        from plone.app.z3cform.widgets.select import AjaxSelectWidget

        portal = self.layer["portal"]
        widget = AjaxSelectWidget(self.request)
        widget.context = portal
        widget.vocabulary = "plone.app.vocabularies.Keywords"
        self.assertEqual(
            widget._base_args()["pattern_options"]["allowNewItems"],
            "false",
        )


def mock_querystring_options(context, querystring_view):
    return {
        "indexOptionsUrl": f"/{querystring_view}",
        "previewURL": "/@@querybuilder_html_results",
        "previewCountURL": "/@@querybuildernumberofresults",
        "patternDateOptions": None,
        "patternAjaxSelectOptions": {"separator": ";"},
        "patternRelateditemsOptions": None,
    }


class QueryStringWidgetTests(unittest.TestCase):
    def setUp(self):
        self.request = TestRequest(environ={"HTTP_ACCEPT_LANGUAGE": "en"})

    def test_converter_toWidgetValue(self):
        from plone.app.z3cform.converters import QueryStringDataConverter

        converter = QueryStringDataConverter(List(), None)
        self.assertEqual(converter.toWidgetValue(None), "[]")
        self.assertEqual(converter.toWidgetValue([]), "[]")

    def test_converter_empty_value(self):
        from plone.app.z3cform.converters import QueryStringDataConverter

        converter = QueryStringDataConverter(List(), None)
        self.assertEqual(converter.toFieldValue(""), None)
        self.assertEqual(converter.toFieldValue("[]"), None)

    @mock.patch(
        "plone.app.z3cform.widgets.querystring.get_querystring_options",
        new=mock_querystring_options,
    )
    def test_widget(self):
        from plone.app.z3cform.widgets.querystring import QueryStringWidget

        widget = QueryStringWidget(self.request)
        self.assertEqual(
            {
                "name": None,
                "value": "",
                "pattern": "querystring",
                "pattern_options": {
                    "indexOptionsUrl": "/@@qsOptions",
                    "previewCountURL": "/@@querybuildernumberofresults",
                    "previewURL": "/@@querybuilder_html_results",
                    "patternAjaxSelectOptions": {"separator": ";"},
                    "patternDateOptions": None,
                    "patternRelateditemsOptions": None,
                },
            },
            widget._base_args(),
        )


class RelatedItemsWidgetIntegrationTests(unittest.TestCase):
    layer = PAZ3CForm_INTEGRATION_TESTING

    def setUp(self):
        self.portal = self.layer["portal"]
        self.request = TestRequest(environ={"HTTP_ACCEPT_LANGUAGE": "en"})
        setRequest(self.request)
        setRoles(self.portal, TEST_USER_ID, ["Manager"])

    def assertDictContainsSubsetReplacement(self, actual, expected):
        """assertDictContainsSubset was removed in Python 3.2, see:
        https://bugs.python.org/issue9424
        To not introduce a forward incompatibility, here is a replacement based
        on: http://stackoverflow.com/a/21058312
        """
        return set(expected.items()).issubset(set(actual.items()))

    def test_related_items_widget(self):
        from plone.app.z3cform.widgets.relateditems import RelatedItemsWidget

        EXPECTED_ROOT_PATH = "/plone"
        EXPECTED_ROOT_URL = "http://nohost/plone"
        EXPECTED_BASE_PATH = "/plone"
        EXPECTED_VOCAB_URL = "http://nohost/plone/@@getVocabulary?name=plone.app.vocabularies.Catalog"  # noqa

        widget = RelatedItemsWidget(self.request)
        widget.context = self.portal
        widget.update()

        result = widget._base_args()

        self.assertEqual(
            EXPECTED_ROOT_PATH,
            result["pattern_options"]["rootPath"],
        )
        self.assertEqual(
            EXPECTED_ROOT_URL,
            result["pattern_options"]["rootUrl"],
        )
        self.assertEqual(
            EXPECTED_BASE_PATH,
            result["pattern_options"]["basePath"],
        )
        self.assertEqual(
            EXPECTED_VOCAB_URL,
            result["pattern_options"]["vocabularyUrl"],
        )

    def test_related_items_widget_nav_root(self):
        from plone.app.layout.navigation.interfaces import INavigationRoot
        from plone.app.z3cform.widgets.relateditems import RelatedItemsWidget

        EXPECTED_ROOT_PATH = "/plone"
        EXPECTED_ROOT_URL = "http://nohost/plone"
        EXPECTED_BASE_PATH = "/plone/subfolder"
        EXPECTED_VOCAB_URL = "http://nohost/plone/@@getVocabulary?name=plone.app.vocabularies.Catalog"  # noqa

        self.portal.invokeFactory("Folder", "subfolder")
        subfolder = self.portal["subfolder"]
        alsoProvides(subfolder, INavigationRoot)

        widget = RelatedItemsWidget(self.request)
        widget.context = subfolder
        widget.update()
        result = widget._base_args()

        self.assertEqual(
            EXPECTED_ROOT_PATH,
            result["pattern_options"]["rootPath"],
        )
        self.assertEqual(
            EXPECTED_ROOT_URL,
            result["pattern_options"]["rootUrl"],
        )
        self.assertEqual(
            EXPECTED_BASE_PATH,
            result["pattern_options"]["basePath"],
        )
        self.assertEqual(
            EXPECTED_VOCAB_URL,
            result["pattern_options"]["vocabularyUrl"],
        )


class IRelationsType(Interface):
    single = RelationChoice(title="Single", required=False, values=[])
    multiple = RelationList(title="Multiple (Relations field)", required=False)


class RelatedItemsWidgetTemplateIntegrationTests(unittest.TestCase):
    layer = PAZ3CForm_INTEGRATION_TESTING

    def setUp(self):
        self.portal = self.layer["portal"]
        self.request = self.layer["request"]
        setRoles(self.portal, TEST_USER_ID, ["Manager"])

    def test_related_items_widget_display_template(self):
        rel_fti = DexterityFTI("RelationsType", schema=IRelationsType.__identifier__)
        self.portal.portal_types._setObject("RelationsType", rel_fti)

        intids = getUtility(IIntIds)

        self.portal.invokeFactory("RelationsType", "source", title="A Source")
        self.portal.invokeFactory("RelationsType", "target", title="A Target")
        self.portal.invokeFactory("Document", "doc", title="A Document")
        source = self.portal["source"]
        target = self.portal["target"]
        doc = self.portal["doc"]

        # Add some relations
        source.single = RelationValue(intids.getId(target))
        source.multiple = [
            RelationValue(intids.getId(target)),
            RelationValue(intids.getId(doc)),
        ]

        # Update relations
        from zope.event import notify
        from zope.lifecycleevent import ObjectModifiedEvent

        notify(ObjectModifiedEvent(source))
        default_view = source.restrictedTraverse("@@view")
        default_view.update()

        single = default_view.w["single"]
        self.assertIsInstance(single, RelatedItemsWidget)
        self.assertTrue(single.value, target.UID())
        items = single.items()
        self.assertIsInstance(items, ContentListing)
        self.assertTrue(items[0].UID, target.UID())

        template = getMultiAdapter(
            (source, self.request, single.form, single.field, single),
            IPageTemplate,
            name=single.mode,
        )
        self.assertTrue(template.filename.endswith("relateditems_display.pt"))
        html = template(single)
        self.assertIn(
            '<span class="contenttype-relationstype state-missing-value url" >A Target</span>',
            html,
        )

        multiple = default_view.w["multiple"]
        self.assertIsInstance(multiple, RelatedItemsWidget)
        self.assertTrue(multiple.value, ";".join([target.UID(), doc.UID()]))
        items = multiple.items()
        self.assertIsInstance(items, ContentListing)
        self.assertTrue(items[0].UID, target.UID())
        self.assertTrue(items[1].UID, doc.UID())

        template = getMultiAdapter(
            (source, self.request, multiple.form, multiple.field, multiple),
            IPageTemplate,
            name=multiple.mode,
        )
        self.assertTrue(template.filename.endswith("relateditems_display.pt"))
        html = template(multiple)
        self.assertIn(
            '<span class="contenttype-relationstype state-missing-value url" >A Target</span>',
            html,
        )
        self.assertIn(
            '<span class="contenttype-document state-missing-value url" >A Document</span>',
            html,
        )


class RelatedItemsWidgetTests(unittest.TestCase):
    def setUp(self):
        self.request = TestRequest(environ={"HTTP_ACCEPT_LANGUAGE": "en"})

    @mock.patch(
        "Products.CMFCore.utils.getToolByName",
        new=Mock(return_value=Mock(return_value="testuser")),
    )
    def test_single_selection(self):
        """The pattern_options value for maximumSelectionSize should
        be 1 when the field only allows a single selection."""
        from plone.app.z3cform.widgets.relateditems import RelatedItemsFieldWidget

        context = Mock(
            absolute_url=lambda: "fake_url", getPhysicalPath=lambda: ["", "site"]
        )
        field = Choice(
            __name__="selectfield",
            values=["one", "two", "three"],
        )
        widget = RelatedItemsFieldWidget(field, self.request)
        widget.context = context
        widget.update()
        base_args = widget._base_args()
        pattern_options = base_args["pattern_options"]
        self.assertEqual(pattern_options.get("maximumSelectionSize", 0), 1)

    @mock.patch(
        "Products.CMFCore.utils.getToolByName",
        new=Mock(return_value=Mock(return_value="testuser")),
    )
    def test_multiple_selection(self):
        """The pattern_options key maximumSelectionSize shouldn't be
        set when the field allows multiple selections"""
        from plone.app.z3cform.widgets.relateditems import RelatedItemsFieldWidget
        from zope.schema.interfaces import ISource
        from zope.schema.vocabulary import VocabularyRegistry

        context = Mock(
            absolute_url=lambda: "fake_url", getPhysicalPath=lambda: ["", "site"]
        )
        field = List(
            __name__="selectfield",
            value_type=Choice(vocabulary="foobar"),
        )
        widget = RelatedItemsFieldWidget(field, self.request)
        widget.context = context

        vocab = Mock()
        alsoProvides(vocab, ISource)
        with mock.patch.object(VocabularyRegistry, "get", return_value=vocab):
            widget.update()
            base_args = widget._base_args()
        patterns_options = base_args["pattern_options"]
        self.assertFalse("maximumSelectionSize" in patterns_options)
        self.assertEqual(
            patterns_options["vocabularyUrl"],
            "/@@getVocabulary?name=foobar&field=selectfield",
        )

    def test_converter_RelationChoice(self):
        from plone.app.z3cform.converters import (
            RelationChoiceRelatedItemsWidgetConverter,
        )

        brain = Mock(getObject=Mock(return_value="obj"))
        portal_catalog = Mock(return_value=[brain])
        widget = Mock()
        converter = RelationChoiceRelatedItemsWidgetConverter(
            TextLine(),
            widget,
        )

        with mock.patch(
            "plone.app.z3cform.converters.IUUID",
            return_value="id",
        ):
            self.assertEqual(converter.toWidgetValue("obj"), "id")
        self.assertEqual(converter.toWidgetValue(None), None)

        with mock.patch(
            "plone.app.z3cform.converters.getToolByName",
            return_value=portal_catalog,
        ):
            self.assertEqual(converter.toFieldValue("id"), "obj")
        self.assertEqual(converter.toFieldValue(None), None)

    def test_converter_RelationList(self):
        from plone.app.z3cform.converters import RelatedItemsDataConverter
        from z3c.relationfield.interfaces import IRelationList

        field = List()
        alsoProvides(field, IRelationList)
        brain1 = Mock(getObject=Mock(return_value="obj1"), UID="id1")
        brain2 = Mock(getObject=Mock(return_value="obj2"), UID="id2")
        portal_catalog = Mock(return_value=[brain1, brain2])
        widget = Mock(separator=";")
        converter = RelatedItemsDataConverter(field, widget)

        self.assertEqual(converter.toWidgetValue(None), None)
        with mock.patch(
            "plone.app.z3cform.converters.IUUID",
            side_effect=["id1", "id2"],
        ):
            self.assertEqual(
                converter.toWidgetValue(["obj1", "obj2"]),
                "id1;id2",
            )

        self.assertEqual(converter.toFieldValue(None), None)
        with mock.patch(
            "plone.app.z3cform.converters.getToolByName",
            return_value=portal_catalog,
        ):
            self.assertEqual(
                converter.toFieldValue("id1;id2"),
                ["obj1", "obj2"],
            )

    def test_converter_List_of_Choice(self):
        from plone.app.z3cform.converters import RelatedItemsDataConverter

        fields = (
            List(),
            List(value_type=TextLine()),
            List(value_type=BytesLine()),
            List(value_type=Choice(values=["one", "two", "three"])),
        )
        for field in fields:
            expected_value_type = getattr(
                field.value_type,
                "_type",
                str,
            )
            if expected_value_type is None:
                expected_value_type = str
            widget = Mock(separator=";")
            converter = RelatedItemsDataConverter(field, widget)

            self.assertEqual(converter.toWidgetValue(None), None)
            self.assertEqual(
                converter.toWidgetValue(["id1", "id2"]),
                "id1;id2",
            )

            self.assertEqual(converter.toFieldValue(None), None)
            if expected_value_type == bytes:
                expected = [b"id1", b"id2"]
            else:
                expected = ["id1", "id2"]
            self.assertEqual(
                converter.toFieldValue("id1;id2"),
                expected,
            )

            self.assertEqual(converter.toFieldValue(None), None)
            self.assertEqual(
                type(converter.toFieldValue("id1;id2")[0]),
                expected_value_type,
            )

    def test_fieldwidget(self):
        from plone.app.z3cform.widgets.relateditems import RelatedItemsFieldWidget
        from plone.app.z3cform.widgets.relateditems import RelatedItemsWidget

        field = Mock(__name__="field", title="", required=True)
        vocabulary = Mock()
        request = Mock()
        widget = RelatedItemsFieldWidget(field, vocabulary, request)
        self.assertTrue(isinstance(widget, RelatedItemsWidget))
        self.assertIs(widget.field, field)
        self.assertIs(widget.request, request)


def add_mock_fti(portal):
    # Fake DX Type
    fti = DexterityFTI("dx_mock")
    portal.portal_types._setObject("dx_mock", fti)
    fti.klass = "plone.dexterity.content.Item"
    fti.schema = "plone.dexterity.tests.schemata.ITestSchema"
    fti.filter_content_types = False
    fti.behaviors = ("plone.app.dexterity.behaviors.metadata.IBasic",)


def _custom_field_widget(field, request):
    from plone.app.z3cform.widgets.select import AjaxSelectWidget

    widget = FieldWidget(field, AjaxSelectWidget(request))
    widget.vocabulary = "plone.app.vocabularies.PortalTypes"
    return widget


class RichTextWidgetTests(unittest.TestCase):
    layer = PAZ3CForm_INTEGRATION_TESTING

    def setUp(self):
        from plone.app.textfield import RichText as RichTextField

        self.portal = self.layer["portal"]
        # since we are using a plone integration layer, we do
        # not need to use the TestRequest class and we do not want
        # to since the tinymce pattern requires a more properly
        # setup zope2 request object to work correctly in tests
        self.request = self.layer["request"]

        class IWithText(Interface):
            text = RichTextField(title="Text")

        self.field = IWithText["text"]

    def test_widget_params(self):
        from plone.app.z3cform.widgets.richtext import RichTextWidget

        widget = FieldWidget(self.field, RichTextWidget(self.request))
        # set the context so we can get tinymce settings
        widget.context = self.portal
        widget.update()
        base_args = widget._base_args()
        self.assertEqual(base_args["name"], "text")
        self.assertEqual(base_args["value"], "")
        self.assertEqual(base_args["pattern"], "tinymce")

        prependToUrl = "/plone/resolveuid/"
        self.assertEqual(
            base_args["pattern_options"]["prependToUrl"],
            prependToUrl,
        )
        self.assertEqual(
            base_args["pattern_options"]["upload"]["relativePath"],
            "@@fileUpload",
        )

    def test_widget_params_different_contexts(self):
        from plone.app.z3cform.widgets.richtext import RichTextWidget

        setRoles(self.portal, TEST_USER_ID, ["Contributor"])

        widget = FieldWidget(self.field, RichTextWidget(self.request))
        self.portal.invokeFactory("Folder", "sub")
        sub = self.portal.sub
        form = Form(sub, self.request)

        # portal context
        widget.context = self.portal
        widget.update()
        base_args = widget._base_args()

        self.assertEqual(
            base_args["pattern_options"]["relatedItems"]["basePath"],
            "/plone",
        )

        # sub context
        widget.context = sub
        widget.update()
        base_args = widget._base_args()

        self.assertEqual(
            base_args["pattern_options"]["relatedItems"]["basePath"],
            "/plone/sub",
        )

        # form context
        widget.context = form
        widget.update()
        base_args = widget._base_args()

        self.assertEqual(
            base_args["pattern_options"]["relatedItems"]["basePath"],
            "/plone/sub",
        )

        # non-contentish context
        widget.context = None
        widget.update()
        base_args = widget._base_args()

        self.assertEqual(
            base_args["pattern_options"]["relatedItems"]["basePath"],
            "/plone",
        )

    def test_widget_values(self):
        from plone.app.textfield.value import RichTextValue
        from plone.app.z3cform.widgets.richtext import RichTextWidget

        widget = FieldWidget(self.field, RichTextWidget(self.request))
        # set the context so we can get tinymce settings
        widget.context = self.portal
        widget.value = RichTextValue("Lorem ipsum \u2026")
        base_args = widget._base_args()
        self.assertEqual(base_args["value"], "Lorem ipsum \u2026")

    def test_unicode_control_characters_value(self):
        # lxml doesn't allow unicode control characters.
        # see
        from plone.app.textfield.value import RichTextValue
        from plone.app.z3cform.widgets.richtext import RichTextWidget

        widget = FieldWidget(self.field, RichTextWidget(self.request))
        # set the context so we can get tinymce settings
        widget.context = self.portal
        widget.value = RichTextValue("Lorem \u0000 ip\u001Fsum\n\u0002 dolorem\r\t")
        widget.mode = "input"
        self.assertIn(
            ">Lorem  ipsum\n dolorem&#13;\t</textarea>",
            widget.render(),
        )

    def _set_mimetypes(self, default="text/html", allowed=("text/html")):
        """Set portal's mimetype settings."""
        if IMarkupSchema:
            registry = getUtility(IRegistry)
            self.settings = registry.forInterface(
                IMarkupSchema,
                prefix="plone",
            )
            self.settings.default_type = default
            self.settings.allowed_types = allowed

    def test_dx_tinymcewidget_single_mimetype(self):
        """A RichTextWidget with only one available mimetype should render the
        pattern class directly on itself.
        """
        if IMarkupSchema:
            # if not, don't run this test
            self._set_mimetypes(allowed=("text/html",))
            from plone.app.z3cform.widgets.richtext import RichTextWidget

            widget = FieldWidget(self.field, RichTextWidget(self.request))
            # set the context so we can get tinymce settings
            widget.context = self.portal
            rendered = widget.render()

            self.assertTrue("<select" not in rendered)
            self.assertTrue("pat-tinymce" in rendered)
            self.assertTrue("data-pat-tinymce" in rendered)

    def test_dx_tinymcewidget_multiple_mimetypes_create(self):
        """A RichTextWidget with multiple available mimetypes should render a
        mimetype selection widget along with the textfield. When there is no
        field value, the default mimetype should be preselected.
        """
        if IMarkupSchema:
            # if not, don't run this test
            self._set_mimetypes(allowed=("text/html", "text/plain"))
            from plone.app.z3cform.widgets.richtext import RichTextWidget

            widget = FieldWidget(self.field, RichTextWidget(self.request))
            # set the context so we can get tinymce settings
            widget.context = self.portal
            rendered = widget.render()

            self.assertTrue("<select" in rendered)
            self.assertTrue("pat-textareamimetypeselector" in rendered)
            self.assertTrue("data-pat-textareamimetypeselector" in rendered)
            self.assertTrue(
                '<option value="text/html" selected="selected">' in rendered
            )
            self.assertTrue("pat-tinymce" not in rendered)

    def test_dx_tinymcewidget_multiple_mimetypes_edit(self):
        """A RichTextWidget with multiple available mimetypes should render a
        mimetype selection widget along with the textfield. When there is
        already a RichTextValue, it's mimetype should be preselected.
        """
        if IMarkupSchema:
            # if not, don't run this test
            self._set_mimetypes(allowed=("text/html", "text/plain"))
            from plone.app.textfield.value import RichTextValue
            from plone.app.z3cform.widgets.richtext import RichTextWidget

            widget = FieldWidget(self.field, RichTextWidget(self.request))
            # set the context so we can get tinymce settings
            widget.context = self.portal
            widget.value = RichTextValue("Hello world", mimeType="text/plain")
            rendered = widget.render()

            self.assertTrue("<select" in rendered)
            self.assertTrue("pat-textareamimetypeselector" in rendered)
            self.assertTrue("data-pat-textareamimetypeselector" in rendered)
            self.assertTrue(
                '<option value="text/plain" selected="selected">' in rendered,
            )
            self.assertTrue("pat-tinymce" not in rendered)

    def test_use_default_editor_value(self):
        """Use dummy utility registered in testing.zcml"""
        if IMarkupSchema:
            # if not, don't run this test
            self._set_mimetypes(allowed=("text/html",))
            registry = getUtility(IRegistry)
            from plone.base.interfaces import IEditingSchema

            proxy = registry.forInterface(IEditingSchema, check=False, prefix="plone")
            proxy.available_editors = ["dummy", "TinyMCE"]
            proxy.default_editor = "dummy"
            from plone.app.z3cform.widgets.richtext import RichTextWidget

            widget = FieldWidget(self.field, RichTextWidget(self.request))
            widget.context = self.portal
            rendered = widget.render()
            self.assertTrue("<p>dummy</p>" in rendered)

            proxy.default_editor = "TinyMCE"
            from plone.app.z3cform.widgets.richtext import RichTextWidget

            widget = FieldWidget(self.field, RichTextWidget(self.request))
            widget.context = self.portal
            rendered = widget.render()
            self.assertTrue("pat-tinymce" in rendered)


class LinkWidgetIntegrationTests(unittest.TestCase):
    layer = PAZ3CForm_INTEGRATION_TESTING

    def setUp(self):
        self.portal = self.layer["portal"]
        self.request = TestRequest(environ={"HTTP_ACCEPT_LANGUAGE": "en"})
        setRequest(self.request)
        setRoles(self.portal, TEST_USER_ID, ["Manager"])

    def test_link_widget__pattern_options(self):
        from plone.app.z3cform.widgets.link import LinkWidget

        widget = LinkWidget(self.request)

        pattern_data = json.loads(widget.pattern_data())
        self.assertEqual(
            pattern_data["vocabularyUrl"],
            "http://nohost/plone/@@getVocabulary?name=plone.app.vocabularies.Catalog",  # noqa
        )
        self.assertEqual(pattern_data["maximumSelectionSize"], 1)

    def test_link_widget__extract_internal(self):
        from plone.app.z3cform.widgets.link import LinkWidget

        widget = LinkWidget(self.request)
        widget.context = self.portal
        widget.name = "testlinkwidget"
        widget.update()

        self.request.form["testlinkwidget.internal"] = "abc"
        self.assertEqual(
            widget.extract(),
            "${portal_url}/resolveuid/abc",
        )

    def test_link_widget__extract_external(self):
        from plone.app.z3cform.widgets.link import LinkWidget

        widget = LinkWidget(self.request)
        widget.context = self.portal
        widget.name = "testlinkwidget"
        widget.update()

        self.request.form["testlinkwidget.external"] = "https://plone.org"
        self.assertEqual(
            widget.extract(),
            "https://plone.org",
        )

    def test_link_widget__extract_email(self):
        from plone.app.z3cform.widgets.link import LinkWidget

        widget = LinkWidget(self.request)
        widget.context = self.portal
        widget.name = "testlinkwidget"
        widget.update()

        self.request.form["testlinkwidget.email"] = "dev@plone.org"
        self.assertEqual(
            widget.extract(),
            "mailto:dev@plone.org",
        )

    def test_link_widget__extract_email_including_mailto(self):
        from plone.app.z3cform.widgets.link import LinkWidget

        widget = LinkWidget(self.request)
        widget.context = self.portal
        widget.name = "testlinkwidget"
        widget.update()

        self.request.form["testlinkwidget.email"] = "mailto:dev@plone.org"
        self.assertEqual(
            widget.extract(),
            "mailto:dev@plone.org",
        )

    def test_link_widget__data_converter(self):
        from plone.app.z3cform.converters import LinkWidgetDataConverter
        from plone.app.z3cform.widgets.link import LinkWidget

        field = TextLine(__name__="linkfield")
        widget = LinkWidget(self.request)
        converter = LinkWidgetDataConverter(field, widget)

        self.portal.invokeFactory("Folder", "test")
        portal_url = self.portal.absolute_url()
        portal_path = "/".join(self.portal.getPhysicalPath())

        # Test empty value
        widget_value = converter.toWidgetValue("")
        self.assertEqual(widget_value["internal"], "")
        self.assertEqual(widget_value["external"], "")
        self.assertEqual(widget_value["email"], "")

        # Test external URLs
        self.assertEqual(
            converter.toWidgetValue("https://plone.org")["external"],
            "https://plone.org",
        )

        # Test relative resolveuid URLs
        self.assertEqual(
            converter.toWidgetValue("/resolveuid/1234")["internal"],
            "1234",
        )

        # Test absolute resolveuid URLs on the same domain
        self.assertEqual(
            converter.toWidgetValue(portal_url + "/resolveuid/1234")[
                "internal"
            ],  # noqa
            "1234",
        )

        # Test absolute resolveuid URLs on a different domain
        self.assertEqual(
            converter.toWidgetValue("http://anyurl/resolveuid/1234")[
                "external"
            ],  # noqa
            "http://anyurl/resolveuid/1234",
        )

        # Test interrnal URL paths
        self.assertEqual(
            converter.toWidgetValue(portal_path + "/test")["internal"],
            IUUID(self.portal.test),
        )

        # Test absolute interrnal URLs
        self.assertEqual(
            converter.toWidgetValue(portal_url + "/test")["internal"],
            IUUID(self.portal.test),
        )

        # Test mail
        self.assertEqual(
            converter.toWidgetValue("mailto:me")["email"],
            "me",
        )

        # Test mail with subject
        self.assertEqual(
            converter.toWidgetValue("mailto:me?subject=jep")["email"],
            "me",
        )
        self.assertEqual(
            converter.toWidgetValue("mailto:me?subject=jep")["email_subject"],
            "jep",
        )


class WidgetCustomizingIntegrationTests(unittest.TestCase):
    layer = PAZ3CForm_INTEGRATION_TESTING

    def test_widget_base_wrapper_css(self):
        class ITestDateSchema(Schema):
            widget("my_date", DateWidget, wrapper_css_class="foo")
            my_date = Date(title="My Date")

        class TestForm(AutoExtensibleForm, EditForm):
            ignoreContext = True
            schema = ITestDateSchema

        render = TestForm(self.layer["portal"], self.layer["request"])
        self.assertIn(
            'empty foo" id="formfield-form-widgets-my_date" data-fieldname="form.widgets.my_date"',
            render(),
        )



from zope.deferredimport import deprecated as deprecated_import


deprecated_import(
    "Import from plone.app.z3cform.converters instead (will be removed in Plone 7)",
    DatetimeWidgetConverter="plone.app.z3cform.converters:DatetimeWidgetConverter",
    DateWidgetConverter="plone.app.z3cform.converters:DateWidgetConverter",
)

deprecated_import(
    "Import from plone.app.z3cform.widgets.base instead (will be removed in Plone 7)",
    BaseWidget="plone.app.z3cform.widgets.base:BaseWidget",
)

deprecated_import(
    "Import from plone.app.z3cform.widgets.datetime instead (will be removed in Plone 7)",
    DateFieldWidget="plone.app.z3cform.widgets.datetime:DateFieldWidget",
    DatetimeFieldWidget="plone.app.z3cform.widgets.datetime:DatetimeFieldWidget",
    DatetimeWidget="plone.app.z3cform.widgets.datetime:DatetimeWidget",
    DateWidget="plone.app.z3cform.widgets.datetime:DateWidget",
    TimeFieldWidget="plone.app.z3cform.widgets.datetime:TimeFieldWidget",
    TimeWidget="plone.app.z3cform.widgets.datetime:TimeWidget",
)

deprecated_import(
    "Import from plone.app.z3cform.widgets.link instead (will be removed in Plone 7)",
    LinkFieldWidget="plone.app.z3cform.widgets.link:LinkFieldWidget",
    LinkWidget="plone.app.z3cform.widgets.link:LinkWidget",
)

deprecated_import(
    "Import from plone.app.z3cform.widgets.querystring instead (will be removed in Plone 7)",
    QueryStringFieldWidget="plone.app.z3cform.widgets.querystring:QueryStringFieldWidget",
    QueryStringWidget="plone.app.z3cform.widgets.querystring:QueryStringWidget",
)

deprecated_import(
    "Import from plone.app.z3cform.widgets.relateditems instead (will be removed in Plone 7)",
    RelatedItemsFieldWidget="plone.app.z3cform.widgets.relateditems:RelatedItemsFieldWidget",
    RelatedItemsWidget="plone.app.z3cform.widgets.relateditems:RelatedItemsWidget",
)

deprecated_import(
    "Import from plone.app.z3cform.widgets.richtext instead (will be removed in Plone 7)",
    RichTextFieldWidget="plone.app.z3cform.widgets.richtext:RichTextFieldWidget",
    RichTextWidget="plone.app.z3cform.widgets.richtext:RichTextWidget",
)

deprecated_import(
    "Import from plone.app.z3cform.widgets.select instead (will be removed in Plone 7)",
    AjaxSelectFieldWidget="plone.app.z3cform.widgets.select:AjaxSelectFieldWidget",
    AjaxSelectWidget="plone.app.z3cform.widgets.select:AjaxSelectWidget",
    SelectFieldWidget="plone.app.z3cform.widgets.select:SelectFieldWidget",
    SelectWidget="plone.app.z3cform.widgets.select:SelectWidget",
)

deprecated_import(
    "Import from plone.app.z3cform.widgets.singlecheckbox instead (will be removed in Plone 7)",
    SingleCheckBoxBoolFieldWidget="plone.app.z3cform.widgets.singlecheckbox:SingleCheckBoxBoolFieldWidget",
    SingleCheckBoxBoolWidget="plone.app.z3cform.widgets.singlecheckbox:SingleCheckBoxBoolWidget",
    SingleCheckBoxWidget="plone.app.z3cform.widgets.singlecheckbox:SingleCheckBoxWidget",
)



import zope.deferredimport


zope.deferredimport.deprecated(
    "Import from plone.z3cform.layout instead.",
    FormWrapper="plone.z3cform.layout:FormWrapper",
    wrap_form="plone.z3cform.layout:wrap_form",
)



from Acquisition import aq_base
from copy import deepcopy
from plone.app.layout.navigation.root import getNavigationRootObject
from Products.CMFCore.interfaces import IContentish
from Products.CMFCore.interfaces import IFolderish
from Products.CMFCore.interfaces import ISiteRoot
from z3c.form.interfaces import IForm
from zope.component import providedBy
from zope.component.hooks import getSite
from zope.globalrequest import getRequest

import urllib


def closest_content(context=None):
    """Try to find a usable context, with increasing aggression"""
    # Normally, we should be given a useful context (e.g the page)
    c = context
    c = _valid_context(c)
    if c is not None:
        return c
    # Subforms (e.g. DataGridField) may not have a context set, find out
    # what page is being published
    c = getattr(getRequest(), "PUBLISHED", None)
    c = _valid_context(c)
    if c is not None:
        return c
    # During widget traversal nothing is being published yet, use getSite()
    c = getSite()
    c = _valid_context(c)
    if c is not None:
        return c
    raise ValueError("Cannot find suitable context to bind to source")


def _valid_context(context):
    """Walk up until finding a content item."""
    # Avoid loops. The object id is used as context may not be hashable
    seen = set()
    while context is not None and id(aq_base(context)) not in seen:
        seen.add(id(aq_base(context)))
        if IContentish.providedBy(context) or IFolderish.providedBy(context):
            return context
        parent = getattr(context, "__parent__", None)
        if parent is None:
            parent = getattr(context, "context", None)
        context = parent

    return None


def call_callables(value, *args, **kwargs):
    """Walk recursively through data structure and call all callables, passing
    the arguments and keyword arguments to it.
    """
    ret = value
    if callable(value):
        ret = value(*args, **kwargs)
    elif isinstance(value, list):
        ret = [call_callables(v, *args, **kwargs) for v in value]
    elif isinstance(value, tuple):
        ret = tuple(call_callables(v, *args, **kwargs) for v in value)
    elif isinstance(value, dict):
        ret = {k: call_callables(v, *args, **kwargs) for k, v in value.items()}
    return ret


def replace_link_variables_by_paths(context, url):
    """Take an `url` and replace the variables "${navigation_root_url}" and
    "${portal_url}" by the corresponding paths. `context` is the acquisition
    context.
    """

    def _replace_variable_by_path(url, variable, obj):
        path = "/".join(obj.getPhysicalPath())
        return url.replace(variable, path)

    if not url:
        return url

    portal_state = context.restrictedTraverse("@@plone_portal_state")

    if "${navigation_root_url}" in url:
        url = _replace_variable_by_path(
            url,
            "${navigation_root_url}",
            portal_state.navigation_root(),
        )

    if "${portal_url}" in url:
        url = _replace_variable_by_path(
            url,
            "${portal_url}",
            portal_state.portal(),
        )

    return url


def is_absolute(url):
    """Return ``True``, if url is an absolute url.
    See: https://stackoverflow.com/a/8357518/1337474
    """
    return bool(urllib.parse.urlparse(url).netloc)


def is_same_domain(url1, url2):
    """Return ``True``, if url1 is on the same protocol and domain than url2."""
    purl1 = urllib.parse.urlparse(url1)
    purl2 = urllib.parse.urlparse(url2)
    return purl1.scheme == purl2.scheme and purl1.netloc == purl2.netloc


def dict_merge(dict_a, dict_b):
    """Helper method which merges two dictionaries.

    Recursively merges dict's. not just simple a['key'] = b['key'], if
    both a and b have a key who's value is a dict then dict_merge is called
    on both values and the result stored in the returned dictionary.

    http://www.xormedia.com/recursively-merge-dictionaries-in-python

    :param dict_a: [required] First dictiornary.
    :type dict_a: dict

    :param dict_b: [required] Second dictiornary.
    :type dict_b: dict

    :returns: Merged dictionary.
    :rtype: dict
    """

    if not isinstance(dict_b, dict):
        return dict_b
    result = deepcopy(dict_a)
    for k, v in dict_b.items():
        if k in result and isinstance(result[k], dict):
            result[k] = dict_merge(result[k], v)
        else:
            result[k] = deepcopy(v)
    return result


def get_widget_form(widget):
    form = getattr(widget, "form", None)
    if getattr(aq_base(form), "parentForm", None) is not None:
        form = form.parentForm
    return form


def get_portal():
    closest_site = getSite()
    if closest_site is not None:
        for potential_portal in closest_site.aq_chain:
            if ISiteRoot in providedBy(potential_portal):
                return potential_portal


def get_portal_url(context):
    portal = get_portal()
    if portal:
        root = getNavigationRootObject(context, portal)
        if root:
            try:
                return root.absolute_url()
            except AttributeError:
                return portal.absolute_url()
        else:
            return portal.absolute_url()
    return ""


def get_context_url(context):
    if IForm.providedBy(context):
        # Use the request URL if we are looking at an addform
        url = context.request.get("URL")
    elif hasattr(context, "absolute_url"):
        url = context.absolute_url
        if callable(url):
            url = url()
    else:
        url = get_portal_url(context)
    return url


# Invalid XML unicode control characters
# NOTE: these control characters are allowed:
# chr(9) = "\t"
# chr(10) = "\n"
# chr(13) = "\r"

_unicode_ctl_chr_map = dict.fromkeys([x for x in range(32) if x not in (9, 10, 13)])


def remove_invalid_xml_characters(txt):
    # remove occurrences of the unicode "control characters"
    # as they are invalid XML characters
    # see https://en.wikipedia.org/wiki/Valid_characters_in_XML and
    # https://en.wikipedia.org/wiki/C0_and_C1_control_codes
    return txt.translate(_unicode_ctl_chr_map)



from lxml import etree
from plone.app.z3cform.utils import remove_invalid_xml_characters

import collections
import json


def el_attrib(name):
    """Helper property methods to get/set/delete element property.

    :param name: [required] Name of the element property.
    :type name: string

    :returns: Property with getter/setter/deletter.
    :rtype: property
    """

    def _get(self):
        if name in self.el.attrib:
            value = self.el.attrib[name]
            if value.strip().startswith("[") or value.strip().startswith("{"):
                value = json.loads(value)
            return value

    def _set(self, value):
        if value is None:
            return
        if isinstance(value, (list, tuple)):
            value = " ".join(value)
        if isinstance(value, (dict, set)):
            value = json.dumps(value)
        if isinstance(value, bytes):
            value = value.decode("utf8")
        self.el.attrib[name] = value

    def _del(self):
        if name in self.el.attrib:
            del self.el.attrib[name]

    return property(_get, _set, _del)


class BaseWidget:
    """Basic patterns widget."""

    _klass_prefix = "pat-"
    klass = el_attrib("class")

    def __init__(self, el, pattern, pattern_options={}):
        """
        :param el: [required] element type (eg. input, div, textarea, a, ...).
        :type el: string

        :param pattern: [required] Pattern name.
        :type pattern: string

        :param pattern_options: Patterns options.
        :type pattern_options: dict
        """

        self.pattern = pattern
        self.el = etree.Element(el)
        if pattern:
            self.klass = self._klass_prefix + pattern
        self.pattern_options = pattern_options

    def update(self):
        """Updating pattern_options in element `data-*` attribute."""
        if self.pattern_options:
            self.el.attrib["data-" + self._klass_prefix + self.pattern] = json.dumps(
                self.pattern_options
            )

    def render(self):
        """Renders the widget

        :returns: Widget's HTML.
        :rtype: string
        """

        self.update()
        return etree.tostring(self.el, encoding=str)


class InputWidget(BaseWidget):
    """Widget with `input` element."""

    type = el_attrib("type")
    value = el_attrib("value")
    name = el_attrib("name")

    def __init__(self, pattern, pattern_options={}, type="text", name=None, value=None):
        """
        :param pattern: [required] Pattern name.
        :type pattern: string

        :param pattern_options: Patterns options.
        :type pattern_options: dict

        :param type: `type` attribute of element.
        :type type: string

        :param name: `name` attribute of element.
        :type name: string

        :param value: `value` attribute of element.
        :type value: string
        """
        super().__init__("input", pattern, pattern_options)
        self.type = type
        if name is not None:
            self.name = name
        if value is not None:
            self.value = value


class SelectWidget(BaseWidget):
    """Widget with `select` element."""

    name = el_attrib("name")
    _multiple = el_attrib("multiple")

    def __init__(
        self,
        pattern,
        pattern_options={},
        items=[],
        name=None,
        value=None,
        multiple=False,
    ):
        """
        :param pattern: [required] Pattern name.
        :type pattern: string

        :param pattern_options: Patterns options.
        :type pattern_options: dict

        :param items: List of value and title pairs which represents possible
                      options to choose from.
        :type items: list

        :param name: `name` attribute of element.
        :type name: string

        :param value: `value` attribute of element.
        :type value: string

        :param multiple: `multiple` attribute of element.
        :type multiple: bool
        """
        super().__init__("select", pattern, pattern_options)
        self.el.text = ""
        self.items = items
        self.multiple = multiple
        if name is not None:
            self.name = name
        if value is not None:
            self.value = value

    def _get_items(self):
        """Get list of possible options.

        :returns: List of value and title pairs.
        :rtype: list
        """
        if self.el.find("optgroup") is not None:
            return collections.OrderedDict(
                (
                    group.attrib["label"],
                    [
                        (option.attrib["value"], option.text)
                        for option in group.iter("option")
                    ],
                )
                for group in self.el.iter("optgroup")
            )
        else:
            return [
                (option.attrib["value"], option.text)
                for option in self.el.iter("option")
            ]

    def _set_items(self, value):
        """Set options for element.

        :param value: List of value and title pairs which represents possible
                      options to choose from.
        :type value: list
        """

        def addOptions(el, options):
            """
            Add <option> elements for each vocab item.
            """
            for token, title in options:
                option = etree.SubElement(el, "option")
                option.attrib["value"] = token
                option.text = title

        if isinstance(value, dict):
            for group_label, options in value.items():
                group = etree.SubElement(self.el, "optgroup")
                group.attrib["label"] = group_label
                addOptions(group, options)
        else:
            for token, title in value:
                option = etree.SubElement(self.el, "option")
                option.attrib["value"] = token
                option.text = title

    def _del_items(self):
        """Removing options from inside of elements."""
        for group in self.el.iter("optgroup"):
            self.el.remove(group)
        for element in self.el.iter("option"):
            self.el.remove(element)

    items = property(_get_items, _set_items, _del_items)

    def _get_value(self):
        """Return selected option(s).

        :returns: Returns list of selected option(s) values.
        :rtype: list
        """
        value = []
        for element in self.el.iter("option"):
            if (
                "selected" in element.attrib
                and element.attrib["selected"] == "selected"
            ):
                value.append(element.attrib["value"])
        return value

    def _set_value(self, value):
        """Select option(s).

        :param value: We are expecting option's value which should be selected.
        :type value: list or string
        """
        if isinstance(value, str):
            value = [value]

        for element in self.el.iter("option"):
            if element.attrib["value"] in value:
                element.attrib["selected"] = "selected"
            elif (
                "selected" in element.attrib
                and element.attrib["selected"] == "selected"
            ):
                del element.attrib["selected"]

    def _del_value(self):
        """Unselect all selected options."""
        for element in self.el.iter("option"):
            if (
                "selected" in element.attrib
                and element.attrib["selected"] == "selected"
            ):
                del element.attrib["selected"]

    value = property(_get_value, _set_value, _del_value)

    def _get_multiple(self):
        """Does element allows multiple items to be selected.

        :returns: `True` if allows multiple elements to be selected, otherwise
                  `False`.
        :rtype: bool
        """
        if self._multiple == "multiple":
            return True
        return False

    def _set_multiple(self, value):
        """Make element accept multiple values.

        :param value: `True` if you want to set element as `multiple`,
                      otherwise `False`
        :type value: bool
        """
        if value:
            self._multiple = "multiple"
        else:
            self._del_multiple()

    def _del_multiple(self):
        """Remove `multiple` attribute from element."""
        del self._multiple

    multiple = property(_get_multiple, _set_multiple, _del_multiple)


class TextareaWidget(BaseWidget):
    """Widget with `textarea` element."""

    name = el_attrib("name")

    def __init__(self, pattern, pattern_options={}, name=None, value=None):
        """
        :param pattern: [required] Pattern name.
        :type pattern: string

        :param pattern_options: Patterns options.
        :type pattern_options: dict

        :param name: `name` attribute of element.
        :type name: string

        :param value: `value` of element.
        :type value: string
        """
        super().__init__("textarea", pattern, pattern_options)
        self.el.text = ""
        if name is not None:
            self.name = name
        if value is not None:
            self.value = value

    def _get_value(self):
        """
        :returns: Value of element.
        :rtype: string
        """
        return self.el.text

    def _set_value(self, value):
        """
        :param value: Set value of element.
        :type value: string
        """
        self.el.text = remove_invalid_xml_characters(value)

    def _del_value(self):
        """Set empty string as value of element."""
        self.el.text = ""

    value = property(_get_value, _set_value, _del_value)



from Acquisition import aq_base
from plone.app.z3cform.utils import call_callables
from z3c.form.widget import Widget


class PatternNotImplemented(Exception):
    """Raised when method/property is not implemented"""


class BaseWidget(Widget):
    """Base widget for z3c.form."""

    pattern = None
    pattern_options = {}
    _adapterValueAttributes = Widget._adapterValueAttributes + ("pattern_options",)

    def _base(self, pattern, pattern_options={}):
        """Base widget class."""
        raise PatternNotImplemented

    def _base_args(self):
        """Method which will calculate _base class arguments.

        Returns (as python dictionary):
            - `pattern`: pattern name
            - `pattern_options`: pattern options

        :returns: Arguments which will be passed to _base
        :rtype: dict
        """
        if self.pattern is None:
            raise PatternNotImplemented("'pattern' option is not provided.")
        return {
            "pattern": self.pattern,
            "pattern_options": self.pattern_options.copy(),
        }

    def render(self):
        """Render widget.

        :returns: Widget's HTML.
        :rtype: string
        """
        if self.mode != "input":
            return super().render()

        _base_args = self._base_args()
        _base_args["pattern_options"] = call_callables(
            _base_args["pattern_options"],
            self.context,
        )

        pattern_widget = self._base(**_base_args)
        if getattr(self, "klass", False):
            pattern_widget.klass = "{} {}".format(
                pattern_widget.klass,
                self.klass,
            )
        return pattern_widget.render()

    def is_subform_widget(self):
        return getattr(aq_base(self.form), "parentForm", None) is not None



from plone.app.z3cform.interfaces import IQueryStringWidget
from plone.app.z3cform.utils import dict_merge
from plone.app.z3cform.utils import get_portal_url
from plone.app.z3cform.widgets.base import BaseWidget
from plone.app.z3cform.widgets.datetime import get_date_options
from plone.app.z3cform.widgets.patterns import InputWidget
from plone.app.z3cform.widgets.relateditems import get_relateditems_options
from z3c.form.browser.text import TextWidget as z3cform_TextWidget
from z3c.form.interfaces import IFieldWidget
from z3c.form.widget import FieldWidget
from zope.globalrequest import getRequest
from zope.interface import implementer
from zope.interface import implementer_only


def get_querystring_options(context, querystring_view):
    portal_url = get_portal_url(context)
    try:
        base_url = context.absolute_url()
    except AttributeError:
        base_url = portal_url
    return {
        "indexOptionsUrl": f"{portal_url}/{querystring_view}",
        "previewURL": "%s/@@querybuilder_html_results" % base_url,
        "previewCountURL": "%s/@@querybuildernumberofresults" % base_url,
        "patternDateOptions": get_date_options(getRequest()),
        "patternAjaxSelectOptions": {"separator": ";"},
        "patternRelateditemsOptions": get_relateditems_options(
            context,
            None,
            ";",
            "plone.app.vocabularies.Catalog",
            "@@getVocabulary",
            "relatedItems",
            include_recently_added=False,
        ),
    }


@implementer_only(IQueryStringWidget)
class QueryStringWidget(BaseWidget, z3cform_TextWidget):
    """QueryString widget for z3c.form."""

    _base = InputWidget

    pattern = "querystring"
    pattern_options = BaseWidget.pattern_options.copy()

    querystring_view = "@@qsOptions"

    def _base_args(self):
        """Method which will calculate _base class arguments.

        Returns (as python dictionary):
            - `pattern`: pattern name
            - `pattern_options`: pattern options
            - `name`: field name
            - `value`: field value

        :returns: Arguments which will be passed to _base
        :rtype: dict
        """
        args = super()._base_args()
        args["name"] = self.name
        args["value"] = self.value

        args.setdefault("pattern_options", {})
        args["pattern_options"] = dict_merge(
            get_querystring_options(self.context, self.querystring_view),
            args["pattern_options"],
        )

        return args


@implementer(IFieldWidget)
def QueryStringFieldWidget(field, request, extra=None):
    if extra is not None:
        request = extra
    return FieldWidget(field, QueryStringWidget(request))



from plone.app.z3cform.tests.layer import PAZ3CForm_INTEGRATION_TESTING
from plone.browserlayer.layer import mark_layer
from plone.testing import layered
from zope.traversing.interfaces import BeforeTraverseEvent

import doctest
import unittest
import zope.component.testing


ROBOT_TEST_LEVEL = 5


class IntegrationTests(unittest.TestCase):
    layer = PAZ3CForm_INTEGRATION_TESTING

    def setUp(self):
        self.portal = self.layer["portal"]
        self.request = self.layer["request"]
        event = BeforeTraverseEvent(self.portal, self.request)
        mark_layer(self.portal, event)

    def test_layer_applied(self):
        from plone.app.z3cform.interfaces import IPloneFormLayer

        self.assertTrue(IPloneFormLayer.providedBy(self.portal.REQUEST))

    def test_default_templates(self):
        form = self.portal.restrictedTraverse("test-form")
        rendered = form()
        # look for something only in the Plone-specific @@ploneform-macros
        self.assertTrue("documentFirstHeading" in rendered)

    def test_content_provider(self):
        form = self.portal.restrictedTraverse("test-form")
        rendered = form()
        self.assertTrue("My test content provider" in rendered)


def test_suite():
    inlineValidationTests = layered(
        doctest.DocFileSuite(
            "inline_validation.rst",
            package="plone.app.z3cform",
            optionflags=(doctest.ELLIPSIS | doctest.NORMALIZE_WHITESPACE),
        ),
        layer=PAZ3CForm_INTEGRATION_TESTING,
    )

    suite = unittest.TestSuite(
        [
            unittest.defaultTestLoader.loadTestsFromTestCase(IntegrationTests),
            doctest.DocFileSuite(
                "wysiwyg/README.rst",
                package="plone.app.z3cform",
                setUp=zope.component.testing.setUp,
                tearDown=zope.component.testing.tearDown,
            ),
            doctest.DocTestSuite(
                "plone.app.z3cform.wysiwyg.widget",
                package="plone.app.z3cform",
                setUp=zope.component.testing.setUp,
                tearDown=zope.component.testing.tearDown,
            ),
            inlineValidationTests,
        ]
    )
    return suite



from plone.dexterity.browser.add import DefaultAddForm
from plone.dexterity.browser.add import DefaultAddView
from plone.dexterity.browser.edit import DefaultEditForm
from plone.dexterity.interfaces import IDexterityEditForm
from plone.z3cform import layout
from Products.Five.browser import BrowserView
from Products.Five.browser.metaconfigure import ViewMixinForTemplates
from z3c.form.error import ErrorViewTemplateFactory
from zope.browserpage.viewpagetemplatefile import ViewPageTemplateFile
from zope.interface import classImplements

import os.path
import plone.app.z3cform
import plone.app.z3cform.interfaces
import plone.z3cform.interfaces
import plone.z3cform.templates
import z3c.form.interfaces


def path(filepart):
    return os.path.join(
        os.path.dirname(plone.app.z3cform.__file__),
        "templates",
        filepart,
    )


# Override the layout wrapper view default template with a more Plone-looking
# one


layout_factory = plone.z3cform.templates.ZopeTwoFormTemplateFactory(
    path("layout.pt"),
    form=plone.z3cform.interfaces.IFormWrapper,
    request=plone.app.z3cform.interfaces.IPloneFormLayer,
)

# Override the form for the standard full-page form rendering

form_factory = plone.z3cform.templates.ZopeTwoFormTemplateFactory(
    path("form.pt"),
    form=z3c.form.interfaces.IForm,
    request=plone.app.z3cform.interfaces.IPloneFormLayer,
)


# The ploneform-macros view
class Macros(BrowserView):
    def __getitem__(self, key):
        return self.index.macros[key]


# The widget rendering templates need to be Zope 3 templates
class RenderWidget(ViewMixinForTemplates, BrowserView):
    index = ViewPageTemplateFile("templates/widget.pt")


class RenderContentProvider(ViewMixinForTemplates, BrowserView):
    index = ViewPageTemplateFile("templates/contentprovider-widget.pt")


ErrorViewTemplate = ErrorViewTemplateFactory(
    os.path.join(os.path.dirname(__file__), "templates/error.pt"), "text/html"
)


class BootstrapActions:
    def updateActions(self):
        super().updateActions()

        for a in self.actions:
            if a == "save":
                self.actions[a].addClass("btn-primary")
            else:
                # this sets 'formnovalidate' attribute in markup
                # to ignore HTML5 validation when clicking buttons other than "save"
                self.actions[a].ignoreRequiredOnValidation = True


# Dexterity Add/Edit Form overrides
class AddForm(BootstrapActions, DefaultAddForm):
    """with bootstrap actions"""


class AddView(DefaultAddView):
    form = AddForm


class EditForm(BootstrapActions, DefaultEditForm):
    """with bootstrap actions"""


EditView = layout.wrap_form(EditForm)
classImplements(EditView, IDexterityEditForm)



from zope.component import adapter
from zope.component.hooks import getSite
from zope.interface import implementer
from zope.interface import implementer_only

import Acquisition
import logging
import warnings
import z3c.form.browser.textarea
import z3c.form.interfaces
import z3c.form.widget
import zope.interface
import zope.schema.interfaces


logger = logging.getLogger("plone.app.z3cform")


class IWysiwygWidget(z3c.form.interfaces.ITextAreaWidget):
    pass


@implementer_only(IWysiwygWidget)
class WysiwygWidget(z3c.form.browser.textarea.TextAreaWidget):
    klass = "kupu-widget"
    value = ""

    def update(self):
        super(z3c.form.browser.textarea.TextAreaWidget, self).update()
        z3c.form.browser.widget.addFieldClass(self)
        # We'll wrap context in the current site *if* it's not already
        # wrapped.  This allows the template to acquire tools with
        # ``context/portal_this`` if context is not wrapped already.
        # Any attempts to satisfy the Kupu template in a less idiotic
        # way failed:
        if getattr(self.form.context, "aq_inner", None) is None:
            self.form.context = Acquisition.ImplicitAcquisitionWrapper(
                self.form.context, getSite()
            )


@adapter(zope.schema.interfaces.IField, z3c.form.interfaces.IFormLayer)
@implementer(z3c.form.interfaces.IFieldWidget)
def WysiwygFieldWidget(field, request):
    """IFieldWidget factory for WysiwygWidget."""
    warnings.warn(
        "plone.app.z3cform.wysiwyg.WysiwygFieldWidget is deprecated and"
        "will be removed in Plone 5.1",
        DeprecationWarning,
    )
    return z3c.form.widget.FieldWidget(field, WysiwygWidget(request))



from plone.app.z3cform.interfaces import ILinkWidget
from plone.base.utils import safe_text
from z3c.form.browser.text import TextWidget as z3cform_TextWidget
from z3c.form.interfaces import IFieldWidget
from z3c.form.interfaces import NO_VALUE
from z3c.form.widget import FieldWidget
from zope.component.hooks import getSite
from zope.interface import implementer
from zope.interface import implementer_only

import json


@implementer_only(ILinkWidget)
class LinkWidget(z3cform_TextWidget):
    """Implementation of enhanced link widget.

    .. note::
        Unlike the others here, this is not a pattern based widget
        and it uses it's own template.
    """

    def pattern_data(self):
        pattern_data = {
            "vocabularyUrl": "{}/@@getVocabulary?name=plone.app.vocabularies.Catalog".format(  # noqa
                getSite().absolute_url(0),
            ),
            "maximumSelectionSize": 1,
        }
        return json.dumps(pattern_data)

    def extract(self, default=NO_VALUE):
        form = self.request.form
        internal = form.get(self.name + ".internal")
        external = form.get(self.name + ".external")
        email = form.get(self.name + ".email")
        if internal:
            url = "${portal_url}/resolveuid/" + internal
        elif email:
            subject = form.get(self.name + ".subject")
            if email[:7] != "mailto:":
                email = "mailto:" + email
            if not subject:
                url = email
            else:
                url = "{email}?subject={subject}".format(
                    email=email,
                    subject=subject,
                )
        else:
            url = external  # the default is `http://` so we land here
        if url:
            self.request.form[self.name] = safe_text(url)
        return super().extract(default=default)


@implementer(IFieldWidget)
def LinkFieldWidget(field, request):
    return FieldWidget(field, LinkWidget(request))



from setuptools import find_packages
from setuptools import setup

import os


def read(*rnames):
    return open(os.path.join(os.path.dirname(__file__), *rnames)).read()


version = "4.2.0"

long_description = (
    read("README.rst")
    + "\n"
    + read("plone", "app", "z3cform", "wysiwyg", "README.rst")
    + "\n"
    + read("plone", "app", "z3cform", "inline_validation.rst")
    + "\n"
    + read("CHANGES.rst")
    + "\n"
)
test_requirements = [
    "plone.app.contenttypes[test]",
    "plone.app.layout",
    "plone.app.testing",
    "plone.autoform",
    "plone.browserlayer",
    "plone.supermodel",
    "plone.testing",
    "zope.annotation",
    "zope.intid",
    "zope.publisher",
]

setup(
    name="plone.app.z3cform",
    version=version,
    description="A collection of widgets, templates and other components "
    "for use with z3c.form and Plone",
    long_description=long_description,
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Web Environment",
        "Framework :: Plone",
        "Framework :: Plone :: 6.0",
        "Framework :: Plone :: Core",
        "Framework :: Zope :: 5",
        "License :: OSI Approved :: GNU General Public License v2 (GPLv2)",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
    ],
    keywords="zope plone form widget template",
    author="Plone Foundation",
    author_email="plone-developers@lists.sourceforge.net",
    url="https://pypi.org/project/plone.app.z3cform",
    license="GPL",
    packages=find_packages(),
    namespace_packages=["plone", "plone.app"],
    include_package_data=True,
    zip_safe=False,
    python_requires=">=3.8",
    install_requires=[
        "plone.app.textfield>=1.3.6",
        "plone.base",
        "plone.protect",
        "plone.registry",
        "plone.schema",
        "plone.uuid",
        "plone.z3cform",
        "Products.GenericSetup",
        "pytz",
        "setuptools",
        "z3c.form >= 4.0",
        "z3c.formwidget.query",
        "zope.deprecation",
        "zope.globalrequest",
        "Zope",
    ],
    extras_require={
        # Until plone.app.z3cform 4.0.2 we only had the 'tests' extra.
        # In 4.0.3 we introduced the 'test' extra.
        # Keep 'tests' for backwards compatibility.
        # Remove it in Plone 7.
        "test": test_requirements,
        "tests": test_requirements,
    },
)



from plone.app.textfield.widget import IRichTextWidget as patextfield_IRichTextWidget
from plone.schema.interfaces import IFormLayer
from z3c.form.interfaces import IRadioWidget
from z3c.form.interfaces import ISelectWidget as IBaseSelectWidget
from z3c.form.interfaces import ISingleCheckBoxWidget
from z3c.form.interfaces import ITextWidget
from zope.interface import Interface
from zope.schema.interfaces import IDate
from zope.schema.interfaces import IDatetime


class IPloneFormLayer(IFormLayer):
    """Request layer installed via browserlayer.xml"""


class IFieldPermissionChecker(Interface):
    """Adapter factory for checking whether a user has permission to
    edit a specific field on a content object.
    """

    def validate(field_name, vocabulary_name=None):
        """Returns True if the current user has permission to edit the
        `field_name` field.  Returns False if the user does not have
        permission.  Raises and AttributeError if the field cannot be
        found.
        """


class IDateField(IDate):
    """Marker interface for the DateField."""


class IDatetimeField(IDatetime):
    """Marker interface for the DatetimeField."""


class IDateWidget(ITextWidget):
    """Marker interface for the DateWidget."""


class IDatetimeWidget(ITextWidget):
    """Marker interface for the DatetimeWidget."""


class ITimeWidget(ITextWidget):
    """Marker interface for the TimeWidget"""


class ISelectWidget(IBaseSelectWidget):
    """Marker interface for the SelectWidget."""


class IAjaxSelectWidget(ITextWidget):
    """Marker interface for the Select2Widget."""


class IQueryStringWidget(ITextWidget):
    """Marker interface for the QueryStringWidget."""


class IRelatedItemsWidget(ITextWidget):
    """Marker interface for the RelatedItemsWidget."""


class IRichTextWidget(patextfield_IRichTextWidget):
    """Marker interface for the TinyMCEWidget."""


class ILinkWidget(ITextWidget):
    """Marker interface for the enhanced link widget."""


class ISingleCheckBoxBoolWidget(ISingleCheckBoxWidget):
    """Marker interface for the SingleCheckboxBoolWidget."""


class IRadioWidget(IRadioWidget):
    """Radio widget."""

    def renderForValue(value):
        """Render a single radio button element for a given value.

        Here the word ``value`` is used in the HTML sense, in other
        words it is a term token.
        """


class IRichTextWidgetInputModeRenderer(Interface):
    """Marker interface to render multiple wysiwyg editors"""



from Acquisition import ImplicitAcquisitionWrapper
from collections import UserDict
from lxml import etree
from plone.app.textfield.value import RichTextValue
from plone.app.textfield.widget import RichTextWidget as patext_RichTextWidget
from plone.app.z3cform.interfaces import IRichTextWidget
from plone.app.z3cform.interfaces import IRichTextWidgetInputModeRenderer
from plone.app.z3cform.utils import closest_content
from plone.app.z3cform.utils import dict_merge
from plone.app.z3cform.widgets.base import BaseWidget
from plone.app.z3cform.widgets.patterns import TextareaWidget
from z3c.form.interfaces import IFieldWidget
from z3c.form.widget import FieldWidget
from zope.component import ComponentLookupError
from zope.component import getMultiAdapter
from zope.component import queryUtility
from zope.interface import implementer
from zope.interface import implementer_only

import json
import logging


logger = logging.getLogger(__name__)


def get_tinymce_options(context, field, request):
    """
    We're just going to be looking up settings from
    plone pattern options
    """
    options = {}
    try:
        pattern_options = getMultiAdapter(
            (context, request, field), name="plone_settings"
        ).tinymce()["data-pat-tinymce"]
        options = json.loads(pattern_options)
    except (ComponentLookupError, AttributeError):
        logger.warning("Can not load tinymce pattern options!", exc_info=True)
    return options


@implementer_only(IRichTextWidget)
class RichTextWidget(BaseWidget, patext_RichTextWidget):
    """TinyMCE widget for z3c.form."""

    _base = TextareaWidget

    pattern_options = BaseWidget.pattern_options.copy()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._pattern = None

    def wrapped_context(self):
        """ "We need to wrap the context to be able to acquire the root
        of the site to get tools, as done in plone.app.textfield"""
        context = self.context
        content = closest_content(context)
        if context.__class__ == dict:
            context = UserDict(self.context)
        return ImplicitAcquisitionWrapper(context, content)

    @property
    def pattern(self):
        """dynamically grab the actual pattern name so it will
        work with custom visual editors"""
        if self._pattern is None:
            self._pattern = self.getWysiwygEditor()
        return self._pattern

    def _base_args(self):
        args = super()._base_args()
        args["name"] = self.name
        value = self.value and self.value.raw or ""
        value = self.request.get(self.name, value)
        args["value"] = value

        args.setdefault("pattern_options", {})
        merged_options = dict_merge(
            get_tinymce_options(
                self.wrapped_context(),
                self.field,
                self.request,
            ),
            args["pattern_options"],
        )
        args["pattern_options"] = merged_options

        return args

    def render(self):
        """Render widget.

        :returns: Widget's HTML.
        :rtype: string
        """
        if self.mode != "display":
            renderer = queryUtility(
                IRichTextWidgetInputModeRenderer,
                name=self.getWysiwygEditor(),
                default=tinymce_richtextwidget_render,
            )
            return renderer(self)

        if not self.value:
            return ""

        if isinstance(self.value, RichTextValue):
            return self.value.output_relative_to(self.context)

        return super().render()

    def render_input_mode(self):
        # MODE "INPUT"
        rendered = ""
        allowed_mime_types = self.allowedMimeTypes()
        if not allowed_mime_types or len(allowed_mime_types) <= 1:
            # Display textarea with default widget
            rendered = super().render()
        else:
            # Let pat-textarea-mimetype-selector choose the widget

            # Initialize the widget without a pattern
            base_args = self._base_args()
            pattern_options = base_args["pattern_options"]
            del base_args["pattern"]
            del base_args["pattern_options"]
            textarea_widget = self._base(None, None, **base_args)
            textarea_widget.klass = "form-control"
            mt_pattern_name = "{}{}".format(
                self._base._klass_prefix,
                "textareamimetypeselector",
            )

            # Initialize mimetype selector pattern
            # TODO: default_mime_type returns 'text/html', regardless of
            # settings. fix in plone.app.textfield
            value_mime_type = (
                self.value.mimeType if self.value else self.field.default_mime_type
            )
            mt_select = etree.Element("select")
            mt_select.attrib["id"] = f"{self.id}_text_format"
            mt_select.attrib["name"] = f"{self.name}.mimeType"
            mt_select.attrib["class"] = f"form-select {mt_pattern_name}"
            mt_select.attrib[f"data-{mt_pattern_name}"] = json.dumps(
                {
                    "textareaName": self.name,
                    "widgets": {
                        "text/html": {  # TODO: currently, we only support
                            # richtext widget config for
                            # 'text/html', no other mimetypes.
                            "pattern": self.pattern,
                            "patternOptions": pattern_options,
                        },
                    },
                },
            )

            # Create a list of allowed mime types
            for mt in allowed_mime_types:
                opt = etree.Element("option")
                opt.attrib["value"] = mt
                if value_mime_type == mt:
                    opt.attrib["selected"] = "selected"
                opt.text = mt
                mt_select.append(opt)

            # Render the combined widget
            rendered = "{}\n{}".format(
                textarea_widget.render(),
                etree.tostring(mt_select, encoding="unicode"),
            )
        return rendered


def tinymce_richtextwidget_render(widget):
    return RichTextWidget.render_input_mode(widget)


@implementer(IFieldWidget)
def RichTextFieldWidget(field, request):
    return FieldWidget(field, RichTextWidget(request))



from Acquisition import aq_parent
from OFS.interfaces import IFolder
from OFS.interfaces import ISimpleItem
from plone.app.contentlisting.interfaces import IContentListing
from plone.app.layout.navigation.root import getNavigationRootObject
from plone.app.z3cform.interfaces import IRelatedItemsWidget
from plone.app.z3cform.utils import dict_merge
from plone.app.z3cform.utils import get_context_url
from plone.app.z3cform.utils import get_widget_form
from plone.app.z3cform.widgets.base import BaseWidget
from plone.app.z3cform.widgets.patterns import InputWidget
from plone.base import PloneMessageFactory as _
from plone.base.utils import get_top_site_from_url
from Products.CMFCore.utils import getToolByName
from z3c.form.browser.text import TextWidget as z3cform_TextWidget
from z3c.form.interfaces import IEditForm
from z3c.form.interfaces import IFieldWidget
from z3c.form.interfaces import IForm
from z3c.form.widget import FieldWidget
from zope.component.hooks import getSite
from zope.globalrequest import getRequest
from zope.interface import implementer
from zope.interface import implementer_only
from zope.schema.interfaces import IChoice
from zope.schema.interfaces import ICollection


def get_relateditems_options(
    context,
    value,
    separator,
    vocabulary_name,
    vocabulary_view,
    field_name=None,
    include_recently_added=True,
):
    if IForm.providedBy(context):
        context = context.context

    request = getRequest()
    site = get_top_site_from_url(context, request)
    options = {
        "separator": separator,
    }
    if not vocabulary_name:
        # we need a vocabulary!
        raise ValueError("RelatedItems needs a vocabulary")
    options["vocabularyUrl"] = "{}/{}?name={}".format(
        get_context_url(site),
        vocabulary_view,
        vocabulary_name,
    )
    if field_name:
        options["vocabularyUrl"] += f"&field={field_name}"
    if value:
        options["initialValues"] = {}
        catalog = False
        if vocabulary_name == "plone.app.vocabularies.Catalog":
            catalog = getToolByName(getSite(), "portal_catalog")
        for value in value.split(separator):
            title = value
            if catalog:
                result = catalog(UID=value)
                title = result[0].Title if result else value
            options["initialValues"][value] = title

    nav_root = getNavigationRootObject(context, site)

    if not ISimpleItem.providedBy(context):
        context = nav_root

    # basePath - start to search/browse in here.
    base_path_context = context
    if not IFolder.providedBy(base_path_context):
        base_path_context = aq_parent(base_path_context)
    if not base_path_context:
        base_path_context = nav_root
    options["basePath"] = "/".join(base_path_context.getPhysicalPath())

    # rootPath - Only display breadcrumb elements deeper than this path.
    options["rootPath"] = "/".join(site.getPhysicalPath()) if site else "/"

    # rootUrl: Visible URL up to the rootPath. This is prepended to the
    # currentPath to generate submission URLs.
    options["rootUrl"] = site.absolute_url() if site else ""

    # contextPath - current edited object. Will not be available to select.
    options["contextPath"] = "/".join(context.getPhysicalPath())

    if base_path_context != nav_root:
        options["favorites"] = [
            {
                "title": _("Current Content"),
                "path": "/".join(base_path_context.getPhysicalPath()),
            },
            {"title": _("Start Page"), "path": "/".join(nav_root.getPhysicalPath())},
        ]

    if include_recently_added:
        # Options for recently used key
        tool = getToolByName(context, "portal_membership")
        user = tool.getAuthenticatedMember()
        options["recentlyUsed"] = False  # Keep that off in Plone 5.1
        options["recentlyUsedKey"] = "relateditems_recentlyused_{}_{}".format(
            field_name or "", user.id
        )  # use string substitution with %s here for automatic str casting.

    return options


@implementer_only(IRelatedItemsWidget)
class RelatedItemsWidget(BaseWidget, z3cform_TextWidget):
    """RelatedItems widget for z3c.form."""

    _base = InputWidget

    pattern = "relateditems"
    pattern_options = BaseWidget.pattern_options.copy()

    separator = ";"
    vocabulary = None
    vocabulary_override = False
    vocabulary_view = "@@getVocabulary"
    orderable = False

    def update(self):
        super().update()
        field = getattr(self, "field", None)
        if ICollection.providedBy(self.field):
            field = self.field.value_type
        if (
            not self.vocabulary
            and field is not None
            and getattr(field, "vocabularyName", None)
        ):
            self.vocabulary = field.vocabularyName
            self.vocabulary_override = True
        else:
            self.vocabulary = "plone.app.vocabularies.Catalog"

    def _base_args(self):
        """Method which will calculate _base class arguments.

        Returns (as python dictionary):
            - `pattern`: pattern name
            - `pattern_options`: pattern options
            - `name`: field name
            - `value`: field value

        :returns: Arguments which will be passed to _base
        :rtype: dict
        """
        args = super()._base_args()

        args["name"] = self.name
        args["value"] = self.value
        args.setdefault("pattern_options", {})

        field = None
        if IChoice.providedBy(self.field):
            args["pattern_options"]["maximumSelectionSize"] = 1
            field = self.field
        elif ICollection.providedBy(self.field):
            field = self.field.value_type

        vocabulary_name = self.vocabulary

        field_name = self.field and self.field.__name__ or None

        context = self.context
        view_context = get_widget_form(self)
        # For EditForms and non-Forms (in tests), the vocabulary is looked
        # up on the context, otherwise on the view
        if IEditForm.providedBy(view_context):
            if self.is_subform_widget():
                view_context = self.form.parentForm.context
            elif not ISimpleItem.providedBy(context):
                view_context = self.form.context
            else:
                view_context = context
        elif not IForm.providedBy(view_context):
            view_context = context
        else:
            pass
            # view_context is defined above already

        root_search_mode = (
            args["pattern_options"].get("mode", None)
            and "basePath" not in args["pattern_options"]
        )

        args["pattern_options"] = dict_merge(
            get_relateditems_options(
                view_context,
                args["value"],
                self.separator,
                vocabulary_name,
                self.vocabulary_view,
                field_name,
            ),
            args["pattern_options"],
        )
        if root_search_mode:
            # Delete default basePath option in search mode, when no basePath
            # was explicitly set.
            del args["pattern_options"]["basePath"]
        if (
            not self.vocabulary_override
            and field
            and getattr(field, "vocabulary", None)
        ):
            # widget vocab takes precedence over field
            form_url = self.request.getURL()
            source_url = "{:s}/++widget++{:s}/@@getSource".format(
                form_url,
                self.name,
            )
            args["pattern_options"]["vocabularyUrl"] = source_url

        return args

    def items(self):
        """Return item for the widget values for the display template

        Query the catalog for the widget-value (uuids) to only display items
        that the user is allowed to see. Accessing the value with e.g.
        getattr(self.context, self.__name__) would yield the items unfiltered.
        Uses IContentListing for easy access to MimeTypeIcon and more.
        """
        results = []
        if not self.value:
            return results
        separator = getattr(self, "separator", ";")
        uuids = self.value.split(separator)

        try:
            catalog = getToolByName(self.context, "portal_catalog")
        except AttributeError:
            catalog = getToolByName(getSite(), "portal_catalog")

        brains = catalog(UID=uuids)
        # restore original order
        results = sorted(brains, key=lambda brain: uuids.index(brain.UID))
        return IContentListing(results)


@implementer(IFieldWidget)
def RelatedItemsFieldWidget(field, request, extra=None):
    if extra is not None:
        request = extra
    return FieldWidget(field, RelatedItemsWidget(request))



from plone.app.vocabularies.terms import TermWithDescription
from plone.app.z3cform.interfaces import IPloneFormLayer
from plone.app.z3cform.interfaces import ISingleCheckBoxBoolWidget
from z3c.form.browser.checkbox import SingleCheckBoxWidget
from z3c.form.interfaces import IFieldWidget
from z3c.form.term import BoolTerms
from z3c.form.term import Terms
from z3c.form.widget import FieldWidget
from zope.component import adapter
from zope.interface import implementer
from zope.interface import implementer_only
from zope.schema.interfaces import IBool
from zope.schema.vocabulary import SimpleTerm
from zope.schema.vocabulary import SimpleVocabulary


@implementer_only(ISingleCheckBoxBoolWidget)
class SingleCheckBoxBoolWidget(SingleCheckBoxWidget):
    """Single Input type checkbox widget implementation."""

    klass = "single-checkbox-bool-widget"

    @property
    def label(self):
        if self.mode == "input":
            return ""
        return getattr(self, "_label", "")

    @label.setter
    def label(self, value):
        self._label = value

    @property
    def description(self):
        if self.mode == "input":
            return ""
        return getattr(self, "_description", "")

    @description.setter
    def description(self, value):
        self._description = value

    def updateTerms(self):
        if self.mode == "input":
            # in input mode use only one checkbox with true
            self.terms = Terms()
            self.terms.terms = SimpleVocabulary(
                (
                    TermWithDescription(
                        True,
                        "selected",
                        getattr(self, "_label", None) or self.field.title,
                        getattr(
                            self,
                            "_description",
                            None,
                        )
                        or self.field.description,
                    ),
                )
            )
            return self.terms
        if not self.terms:
            self.terms = Terms()
            self.terms.terms = SimpleVocabulary(
                [
                    SimpleTerm(*args)
                    for args in [
                        (True, "selected", BoolTerms.trueLabel),
                        (False, "unselected", BoolTerms.falseLabel),
                    ]
                ],
            )
        return self.terms

    @property
    def items(self):
        result = super().items
        for record in result:
            term = self.terms.terms.getTermByToken(record["value"])
            record["description"] = getattr(term, "description", "")
            record["required"] = self.required
        return result


@adapter(IBool, IPloneFormLayer)
@implementer(IFieldWidget)
def SingleCheckBoxBoolFieldWidget(field, request):
    """IFieldWidget factory for CheckBoxWidget."""
    return FieldWidget(field, SingleCheckBoxBoolWidget(request))



from plone.app.testing import PLONE_FIXTURE
from plone.app.testing import PloneSandboxLayer
from plone.app.testing.layers import IntegrationTesting
from plone.testing import z2


class PAZ3CFormLayer(PloneSandboxLayer):
    defaultBases = (PLONE_FIXTURE,)

    def setUpZope(self, app, configurationContext):
        z2.installProduct(app, "Products.DateRecurringIndex")

        import plone.app.contenttypes

        self.loadZCML(package=plone.app.contenttypes, context=configurationContext)

        import plone.app.z3cform

        self.loadZCML(package=plone.app.z3cform, context=configurationContext)
        import plone.app.z3cform.tests

        self.loadZCML(
            name="testing.zcml",
            package=plone.app.z3cform.tests,
            context=configurationContext,
        )

    def setUpPloneSite(self, portal):
        self.applyProfile(portal, "plone.app.contenttypes:default")
        self.applyProfile(portal, "plone.app.z3cform:default")


PAZ3CForm_FIXTURE = PAZ3CFormLayer()
PAZ3CForm_INTEGRATION_TESTING = IntegrationTesting(
    bases=(PAZ3CForm_FIXTURE,),
    name="plone.app.z3cform:Integration",
)
